from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Optional

import typer
from rich.progress import Progress

from .. import data as d
from ..api import ProtectApiClient
from ..cli import base
from ..exceptions import NvrError
from ..utils import local_datetime

app = typer.Typer(rich_markup_mode="rich")

ARG_EVENT_ID = typer.Argument(None, help="ID of camera to select for subcommands")
OPTION_START = typer.Option(None, "-s", "--start")
OPTION_END = typer.Option(None, "-e", "--end")
OPTION_LIMIT = typer.Option(None, "-l", "--limit")
OPTION_OFFSET = typer.Option(None, "-o", "--offet")
OPTION_TYPES = typer.Option(None, "-t", "--type")
OPTION_SMART_TYPES = typer.Option(
    None,
    "-d",
    "--smart-detect",
    help="If provided, will only return smartDetectZone events",
)


@dataclass
class EventContext(base.CliContext):
    events: dict[str, d.Event] | None = None
    event: d.Event | None = None


ALL_COMMANDS: dict[str, Callable[..., None]] = {}


@app.callback(invoke_without_command=True)
def main(
    ctx: typer.Context,
    event_id: Optional[str] = ARG_EVENT_ID,
    start: Optional[datetime] = OPTION_START,
    end: Optional[datetime] = OPTION_END,
    limit: Optional[int] = OPTION_LIMIT,
    offset: Optional[int] = OPTION_OFFSET,
    types: Optional[list[d.EventType]] = OPTION_TYPES,
    smart_types: Optional[list[d.SmartDetectObjectType]] = OPTION_SMART_TYPES,
) -> None:
    """
    Events CLI.

    Returns list of events from the last 24 hours without any arguments passed.
    """
    protect: ProtectApiClient = ctx.obj.protect
    context = EventContext(
        protect=ctx.obj.protect,
        event=None,
        events=None,
        output_format=ctx.obj.output_format,
    )
    ctx.obj = context

    if event_id is not None and event_id not in ALL_COMMANDS:
        try:
            ctx.obj.event = base.run(ctx, protect.get_event(event_id))
        except NvrError as err:
            typer.secho("Invalid event ID", fg="red")
            raise typer.Exit(1) from err

    if not ctx.invoked_subcommand:
        if ctx.obj.event is not None:
            base.print_unifi_obj(ctx.obj.event, ctx.obj.output_format)
            return

        if types is not None and len(types) == 0:
            types = None
        if smart_types is not None and len(smart_types) == 0:
            smart_types = None
        events = base.run(
            ctx,
            protect.get_events(
                start=start,
                end=end,
                limit=limit,
                offset=offset,
                types=types,
                smart_detect_types=smart_types,
            ),
        )
        ctx.obj.events = {}
        for event in events:
            ctx.obj.events[event.id] = event

        if event_id in ALL_COMMANDS:
            ctx.invoke(ALL_COMMANDS[event_id], ctx)
            return

        base.print_unifi_dict(ctx.obj.events)


def require_event_id(ctx: typer.Context) -> None:
    """Requires event ID in context"""
    if ctx.obj.event is None:
        typer.secho("Requires a valid event ID to be selected")
        raise typer.Exit(1)


def require_no_event_id(ctx: typer.Context) -> None:
    """Requires no device ID in context"""
    if ctx.obj.event is not None or ctx.obj.events is None:
        typer.secho("Requires no event ID to be selected")
        raise typer.Exit(1)


@app.command()
def list_ids(ctx: typer.Context) -> None:
    """
    Prints list of "id type timestamp" for each event.

    Timestamps dispalyed in your locale timezone. If it is not configured
    correctly, it will default to UTC. You can override your timezone with
    the TZ environment variable.
    """
    require_no_event_id(ctx)
    objs: dict[str, d.Event] = ctx.obj.events
    to_print: list[tuple[str, str, datetime]] = []
    longest_event = 0
    for obj in objs.values():
        event_type = obj.type.value
        if event_type in {
            d.EventType.SMART_DETECT.value,
            d.EventType.SMART_DETECT_LINE.value,
        }:
            event_type = f"{event_type}[{','.join(obj.smart_detect_types)}]"
        longest_event = max(len(event_type), longest_event)
        dt = obj.timestamp or obj.start
        dt = local_datetime(dt)

        to_print.append((obj.id, event_type, dt))

    if ctx.obj.output_format == base.OutputFormatEnum.JSON:
        base.json_output(to_print)
    else:
        for item in to_print:
            typer.echo(f"{item[0]}\t{item[1]:{longest_event}}\t{item[2]}")


ALL_COMMANDS["list-ids"] = list_ids


@app.command()
def save_thumbnail(
    ctx: typer.Context,
    output_path: Path = typer.Argument(..., help="JPEG format"),
) -> None:
    """
    Saves thumbnail for event.

    Only for ring, motion and smartDetectZone events.
    """
    require_event_id(ctx)
    event: d.Event = ctx.obj.event

    thumbnail = base.run(ctx, event.get_thumbnail())
    if thumbnail is None:
        typer.secho("Could not get thumbnail", fg="red")
        raise typer.Exit(1)

    Path(output_path).write_bytes(thumbnail)


@app.command()
def save_animated_thumbnail(
    ctx: typer.Context,
    output_path: Path = typer.Argument(..., help="GIF format"),
) -> None:
    """
    Saves animated thumbnail for event.

    Only for ring, motion and smartDetectZone events.
    """
    require_event_id(ctx)
    event: d.Event = ctx.obj.event

    thumbnail = base.run(ctx, event.get_animated_thumbnail())
    if thumbnail is None:
        typer.secho("Could not get thumbnail", fg="red")
        raise typer.Exit(1)

    Path(output_path).write_bytes(thumbnail)


@app.command()
def save_heatmap(
    ctx: typer.Context,
    output_path: Path = typer.Argument(..., help="PNG format"),
) -> None:
    """
    Saves heatmap for event.

    Only motion events have heatmaps.
    """
    require_event_id(ctx)
    event: d.Event = ctx.obj.event

    heatmap = base.run(ctx, event.get_heatmap())
    if heatmap is None:
        typer.secho("Could not get heatmap", fg="red")
        raise typer.Exit(1)

    Path(output_path).write_bytes(heatmap)


@app.command()
def save_video(
    ctx: typer.Context,
    output_path: Path = typer.Argument(..., help="MP4 format"),
    channel: int = typer.Option(
        0,
        "-c",
        "--channel",
        min=0,
        max=3,
        help="0 = High, 1 = Medium, 2 = Low, 3 = Package",
    ),
) -> None:
    """
    Exports video for event.

    Only for ring, motion and smartDetectZone events.
    """
    require_event_id(ctx)
    event: d.Event = ctx.obj.event

    with Progress() as pb:
        task_id = pb.add_task("(1/2) Exporting", total=100)

        async def callback(step: int, current: int, total: int) -> None:
            pb.update(
                task_id,
                total=total,
                completed=current,
                description="(2/2) Downloading",
            )

        base.run(
            ctx,
            event.get_video(
                channel,
                output_file=output_path,
                progress_callback=callback,
            ),
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

import typer

from ..api import ProtectApiClient
from ..cli import base
from ..data import MountType, Sensor

app = typer.Typer(rich_markup_mode="rich")

ARG_DEVICE_ID = typer.Argument(None, help="ID of sensor to select for subcommands")


@dataclass
class SensorContext(base.CliContext):
    devices: dict[str, Sensor]
    device: Sensor | None = None


ALL_COMMANDS, DEVICE_COMMANDS = base.init_common_commands(app)


@app.callback(invoke_without_command=True)
def main(ctx: typer.Context, device_id: Optional[str] = ARG_DEVICE_ID) -> None:
    """
    Sensors device CLI.

    Returns full list of Sensors without any arguments passed.
    """
    protect: ProtectApiClient = ctx.obj.protect
    context = SensorContext(
        protect=ctx.obj.protect,
        device=None,
        devices=protect.bootstrap.sensors,
        output_format=ctx.obj.output_format,
    )
    ctx.obj = context

    if device_id is not None and device_id not in ALL_COMMANDS:
        if (device := protect.bootstrap.sensors.get(device_id)) is None:
            typer.secho("Invalid sensor ID", fg="red")
            raise typer.Exit(1)
        ctx.obj.device = device

    if not ctx.invoked_subcommand:
        if device_id in ALL_COMMANDS:
            ctx.invoke(ALL_COMMANDS[device_id], ctx)
            return

        if ctx.obj.device is not None:
            base.print_unifi_obj(ctx.obj.device, ctx.obj.output_format)
            return

        base.print_unifi_dict(ctx.obj.devices)


@app.command()
def camera(ctx: typer.Context, camera_id: Optional[str] = typer.Argument(None)) -> None:
    """Returns or sets tha paired camera for a sensor."""
    base.require_device_id(ctx)
    obj: Sensor = ctx.obj.device

    if camera_id is None:
        base.print_unifi_obj(obj.camera, ctx.obj.output_format)
    else:
        protect: ProtectApiClient = ctx.obj.protect
        if (camera_obj := protect.bootstrap.cameras.get(camera_id)) is None:
            typer.secho("Invalid camera ID")
            raise typer.Exit(1)
        base.run(ctx, obj.set_paired_camera(camera_obj))


@app.command()
def is_tampering_detected(ctx: typer.Context) -> None:
    """Returns if tampering is detected for sensor"""
    base.require_device_id(ctx)
    obj: Sensor = ctx.obj.device
    base.json_output(obj.is_tampering_detected)


@app.command()
def is_alarm_detected(ctx: typer.Context) -> None:
    """Returns if alarm is detected for sensor"""
    base.require_device_id(ctx)
    obj: Sensor = ctx.obj.device
    base.json_output(obj.is_alarm_detected)


@app.command()
def is_contact_enabled(ctx: typer.Context) -> None:
    """Returns if contact sensor is enabled for sensor"""
    base.require_device_id(ctx)
    obj: Sensor = ctx.obj.device
    base.json_output(obj.is_contact_sensor_enabled)


@app.command()
def is_motion_enabled(ctx: typer.Context) -> None:
    """Returns if motion sensor is enabled for sensor"""
    base.require_device_id(ctx)
    obj: Sensor = ctx.obj.device
    base.json_output(obj.is_contact_sensor_enabled)


@app.command()
def is_alarm_enabled(ctx: typer.Context) -> None:
    """Returns if alarm sensor is enabled for sensor"""
    base.require_device_id(ctx)
    obj: Sensor = ctx.obj.device
    base.json_output(obj.is_alarm_sensor_enabled)


@app.command()
def is_light_enabled(ctx: typer.Context) -> None:
    """Returns if light sensor is enabled for sensor"""
    base.require_device_id(ctx)
    obj: Sensor = ctx.obj.device
    base.json_output(obj.is_light_sensor_enabled)


@app.command()
def is_temperature_enabled(ctx: typer.Context) -> None:
    """Returns if temperature sensor is enabled for sensor"""
    base.require_device_id(ctx)
    obj: Sensor = ctx.obj.device
    base.json_output(obj.is_temperature_sensor_enabled)


@app.command()
def is_humidity_enabled(ctx: typer.Context) -> None:
    """Returns if humidity sensor is enabled for sensor"""
    base.require_device_id(ctx)
    obj: Sensor = ctx.obj.device
    base.json_output(obj.is_humidity_sensor_enabled)


@app.command()
def set_status_light(ctx: typer.Context, enabled: bool) -> None:
    """Sets status light for sensor device."""
    base.require_device_id(ctx)
    obj: Sensor = ctx.obj.device

    base.run(ctx, obj.set_status_light(enabled))


@app.command()
def set_mount_type(ctx: typer.Context, mount_type: MountType) -> None:
    """Sets mount type for sensor device."""
    base.require_device_id(ctx)
    obj: Sensor = ctx.obj.device

    base.run(ctx, obj.set_mount_type(mount_type))


@app.command()
def set_motion(ctx: typer.Context, enabled: bool) -> None:
    """Sets motion sensor status for sensor device."""
    base.require_device_id(ctx)
    obj: Sensor = ctx.obj.device

    base.run(ctx, obj.set_motion_status(enabled))


@app.command()
def set_temperature(ctx: typer.Context, enabled: bool) -> None:
    """Sets temperature sensor status for sensor device."""
    base.require_device_id(ctx)
    obj: Sensor = ctx.obj.device

    base.run(ctx, obj.set_temperature_status(enabled))


@app.command()
def set_humidity(ctx: typer.Context, enabled: bool) -> None:
    """Sets humidity sensor status for sensor device."""
    base.require_device_id(ctx)
    obj: Sensor = ctx.obj.device

    base.run(ctx, obj.set_humidity_status(enabled))


@app.command()
def set_light(ctx: typer.Context, enabled: bool) -> None:
    """Sets light sensor status for sensor device."""
    base.require_device_id(ctx)
    obj: Sensor = ctx.obj.device

    base.run(ctx, obj.set_light_status(enabled))


@app.command()
def set_alarm(ctx: typer.Context, enabled: bool) -> None:
    """Sets alarm sensor status for sensor device."""
    base.require_device_id(ctx)
    obj: Sensor = ctx.obj.device

    base.run(ctx, obj.set_alarm_status(enabled))


@app.command()
def set_motion_sensitivity(
    ctx: typer.Context,
    sensitivity: int = typer.Argument(..., min=0, max=100),
) -> None:
    """Sets motion sensitivity for the sensor."""
    base.require_device_id(ctx)
    obj: Sensor = ctx.obj.device

    base.run(ctx, obj.set_motion_sensitivity(sensitivity))


@app.command()
def set_temperature_range(
    ctx: typer.Context,
    low: float = typer.Argument(..., min=0, max=44),
    high: float = typer.Argument(..., min=1, max=45),
) -> None:
    """Sets temperature safe range (in °C). Anything out side of range will trigger event."""
    base.require_device_id(ctx)
    obj: Sensor = ctx.obj.device

    base.run(ctx, obj.set_temperature_safe_range(low, high))


@app.command()
def set_humidity_range(
    ctx: typer.Context,
    low: float = typer.Argument(..., min=1, max=98),
    high: float = typer.Argument(..., min=2, max=99),
) -> None:
    """Sets humidity safe range (in relative % humidity). Anything out side of range will trigger event."""
    base.require_device_id(ctx)
    obj: Sensor = ctx.obj.device

    base.run(ctx, obj.set_humidity_safe_range(low, high))


@app.command()
def set_light_range(
    ctx: typer.Context,
    low: float = typer.Argument(..., min=1, max=999),
    high: float = typer.Argument(..., min=2, max=1000),
) -> None:
    """Sets light safe range (in lux). Anything out side of range will trigger event."""
    base.require_device_id(ctx)
    obj: Sensor = ctx.obj.device

    base.run(ctx, obj.set_light_safe_range(low, high))


@app.command()
def remove_temperature_range(ctx: typer.Context) -> None:
    """Removes temperature safe ranges so events will no longer fire."""
    base.require_device_id(ctx)
    obj: Sensor = ctx.obj.device

    base.run(ctx, obj.remove_temperature_safe_range())


@app.command()
def remove_humidity_range(ctx: typer.Context) -> None:
    """Removes humidity safe ranges so events will no longer fire."""
    base.require_device_id(ctx)
    obj: Sensor = ctx.obj.device

    base.run(ctx, obj.remove_humidity_safe_range())


@app.command()
def remove_light_range(ctx: typer.Context) -> None:
    """Removes light safe ranges so events will no longer fire."""
    base.require_device_id(ctx)
    obj: Sensor = ctx.obj.device

    base.run(ctx, obj.remove_light_safe_range())



from __future__ import annotations

import enum
from collections.abc import Callable, Coroutine
from functools import cache
from typing import Any, Literal, Optional, TypeVar, Union

from packaging.version import Version as BaseVersion
from pydantic.v1 import BaseModel, ConstrainedInt
from pydantic.v1.color import Color as BaseColor
from pydantic.v1.types import ConstrainedFloat, ConstrainedStr

from .._compat import cached_property

KT = TypeVar("KT")
VT = TypeVar("VT")


DEFAULT = "DEFAULT_VALUE"
DEFAULT_TYPE = Literal["DEFAULT_VALUE"]
EventCategories = Literal[
    "critical",
    "update",
    "admin",
    "ring",
    "motion",
    "smart",
    "iot",
]

ProgressCallback = Callable[[int, int, int], Coroutine[Any, Any, None]]
IteratorCallback = Callable[[int, Optional[bytes]], Coroutine[Any, Any, None]]


class FixSizeOrderedDict(dict[KT, VT]):
    """A fixed size ordered dict."""

    def __init__(self, *args: Any, max_size: int = 0, **kwargs: Any) -> None:
        """Create the FixSizeOrderedDict."""
        self._max_size = max_size
        super().__init__(*args, **kwargs)

    def __setitem__(self, key: KT, value: VT) -> None:
        """Set an update up to the max size."""
        dict.__setitem__(self, key, value)
        if self._max_size > 0 and len(self) > 0 and len(self) > self._max_size:
            del self[next(iter(self))]


class ValuesEnumMixin:
    _values: list[str] | None = None
    _values_normalized: dict[str, str] | None = None

    @classmethod
    @cache
    def from_string(cls, value: str) -> Any:
        return cls(value)  # type: ignore[call-arg]

    @classmethod
    @cache
    def values(cls) -> list[str]:
        if cls._values is None:
            cls._values = [e.value for e in cls]  # type: ignore[attr-defined]
        return cls._values

    @classmethod
    @cache
    def values_set(cls) -> set[str]:
        return set(cls.values())

    @classmethod
    def _missing_(cls, value: Any) -> Any | None:
        if cls._values_normalized is None:
            cls._values_normalized = {e.value.lower(): e for e in cls}  # type: ignore[attr-defined]

        value_normal = value
        if isinstance(value, str):
            value_normal = value.lower()
        return cls._values_normalized.get(value_normal)


class UnknownValuesEnumMixin(ValuesEnumMixin):
    @classmethod
    def _missing_(cls, value: Any) -> Any | None:
        # value always set in superclass _missing
        return super()._missing_(value) or cls._values_normalized.get("unknown")  # type: ignore[union-attr]


@enum.unique
class ModelType(str, UnknownValuesEnumMixin, enum.Enum):
    CAMERA = "camera"
    CLOUD_IDENTITY = "cloudIdentity"
    EVENT = "event"
    GROUP = "group"
    LIGHT = "light"
    LIVEVIEW = "liveview"
    NVR = "nvr"
    USER = "user"
    USER_LOCATION = "userLocation"
    VIEWPORT = "viewer"
    BRIDGE = "bridge"
    SENSOR = "sensor"
    DOORLOCK = "doorlock"
    SCHEDULE = "schedule"
    CHIME = "chime"
    DEVICE_GROUP = "deviceGroup"
    RECORDING_SCHEDULE = "recordingSchedule"
    UNKNOWN = "unknown"

    bootstrap_model_types: tuple[ModelType, ...]
    bootstrap_models: tuple[str, ...]
    bootstrap_models_set: set[str]
    bootstrap_models_types_set: set[ModelType]
    bootstrap_models_types_and_event_set: set[ModelType]

    @cached_property
    def devices_key(self) -> str:
        """Return the devices key."""
        return f"{self.value}s"

    @cached_property
    def name(self) -> str:
        """Return the name."""
        return self._name_

    @cached_property
    def value(self) -> str:
        """Return the value."""
        return self._value_

    @classmethod
    @cache
    def from_string(cls, value: str) -> ModelType:
        return cls(value)

    @classmethod
    def _bootstrap_model_types(cls) -> tuple[ModelType, ...]:
        """Return the bootstrap models as a tuple."""
        # TODO:
        # legacyUFV
        # display
        return (
            ModelType.CAMERA,
            ModelType.USER,
            ModelType.GROUP,
            ModelType.LIVEVIEW,
            ModelType.VIEWPORT,
            ModelType.LIGHT,
            ModelType.BRIDGE,
            ModelType.SENSOR,
            ModelType.DOORLOCK,
            ModelType.CHIME,
        )

    @classmethod
    def _bootstrap_models(cls) -> tuple[str, ...]:
        """Return the bootstrap models strings as a tuple."""
        return tuple(
            model_type.value for model_type in ModelType._bootstrap_model_types()
        )

    @classmethod
    def _bootstrap_models_set(cls) -> set[str]:
        """Return the set of bootstrap models strings as a set."""
        return set(ModelType._bootstrap_models())

    @classmethod
    def _bootstrap_models_types_set(cls) -> set[ModelType]:
        """Return the set of bootstrap models as a set."""
        return set(ModelType._bootstrap_model_types())

    @classmethod
    def _bootstrap_models_types_and_event_set(cls) -> set[ModelType]:
        """Return the set of bootstrap models and the event model as a set."""
        return ModelType._bootstrap_models_types_set() | {ModelType.EVENT}

    def _immutable(self, name: str, value: Any) -> None:
        raise AttributeError("Cannot modify ModelType")


ModelType.bootstrap_model_types = ModelType._bootstrap_model_types()
ModelType.bootstrap_models = ModelType._bootstrap_models()
ModelType.bootstrap_models_set = ModelType._bootstrap_models_set()
ModelType.bootstrap_models_types_set = ModelType._bootstrap_models_types_set()
ModelType.bootstrap_models_types_and_event_set = (
    ModelType._bootstrap_models_types_and_event_set()
)
ModelType.__setattr__ = ModelType._immutable  # type: ignore[method-assign, assignment]


@enum.unique
class EventType(str, ValuesEnumMixin, enum.Enum):
    DISCONNECT = "disconnect"
    FACTORY_RESET = "factoryReset"
    PROVISION = "provision"
    UPDATE = "update"
    CAMERA_POWER_CYCLE = "cameraPowerCycling"
    RING = "ring"
    DOOR_ACCESS = "doorAccess"
    RESOLUTION_LOWERED = "resolutionLowered"
    POOR_CONNECTION = "poorConnection"
    STREAM_RECOVERY = "streamRecovery"
    MOTION = "motion"
    NFC_CARD_SCANNED = "nfcCardScanned"
    FINGERPRINT_IDENTIFIED = "fingerprintIdentified"
    RECORDING_DELETED = "recordingDeleted"
    SMART_AUDIO_DETECT = "smartAudioDetect"
    SMART_DETECT = "smartDetectZone"
    SMART_DETECT_LINE = "smartDetectLine"
    NO_SCHEDULE = "nonScheduledRecording"
    RECORDING_MODE_CHANGED = "recordingModeChanged"
    HOTPLUG = "hotplug"
    FACE_GROUP_DETECTED = "faceGroupDetected"
    CONSOLIDATED_RESOLUTION_LOWERED = "consolidatedResolutionLowered"
    CONSOLIDATED_POOR_CONNECTION = "consolidatedPoorConnection"
    CAMERA_CONNECTED = "cameraConnected"
    CAMERA_REBOOTED = "cameraRebooted"
    CAMERA_DISCONNECTED = "cameraDisconnected"
    # ---
    INSTALLED_DISK = "installed"
    CORRUPTED_DB_RECOVERED = "corruptedDbRecovered"
    OFFLINE = "offline"
    OFF = "off"
    REBOOT = "reboot"
    FIRMWARE_UPDATE = "fwUpdate"
    APP_UPDATE = "applicationUpdate"
    APPLICATION_UPDATABLE = "applicationUpdatable"
    ACCESS = "access"
    DRIVE_FAILED = "driveFailed"
    CAMERA_UTILIZATION_LIMIT_REACHED = "cameraUtilizationLimitReached"
    CAMERA_UTILIZATION_LIMIT_EXCEEDED = "cameraUtilizationLimitExceeded"
    DRIVE_SLOW = "driveSlow"
    GLOBAL_RECORDING_MODE_CHANGED = "globalRecordingModeChanged"
    NVR_SETTINGS_CHANGED = "nvrSettingsChanged"
    # ---
    UNADOPTED_DEVICE_DISCOVERED = "unadoptedDeviceDiscovered"
    MULTIPLE_UNADOPTED_DEVICE_DISCOVERED = "multipleUnadoptedDeviceDiscovered"
    DEVICE_ADOPTED = "deviceAdopted"
    DEVICE_UNADOPTED = "deviceUnadopted"
    UVF_DISCOVERED = "ufvDiscovered"
    DEVICE_PASSWORD_UPDATE = "devicesPasswordUpdated"  # noqa: S105
    DEVICE_UPDATABLE = "deviceUpdatable"
    MULTIPLE_DEVICE_UPDATABLE = "multipleDeviceUpdatable"
    DEVICE_CONNECTED = "deviceConnected"
    DEVICE_REBOOTED = "deviceRebooted"
    DEVICE_DISCONNECTED = "deviceDisconnected"
    NETWORK_DEVICE_OFFLINE = "networkDeviceOffline"
    # ---
    USER_LEFT = "userLeft"
    USER_ARRIVED = "userArrived"
    VIDEO_EXPORTED = "videoExported"
    MIC_DISABLED = "microphoneDisabled"
    VIDEO_DELETED = "videoDeleted"
    SCHEDULE_CHANGED = "recordingScheduleChanged"
    # ---
    MOTION_SENSOR = "sensorMotion"
    SENSOR_OPENED = "sensorOpened"
    SENSOR_CLOSED = "sensorClosed"
    SENSOR_ALARM = "sensorAlarm"
    SENSOR_EXTREME_VALUE = "sensorExtremeValues"
    SENSOR_WATER_LEAK = "sensorWaterLeak"
    SENSOR_BATTERY_LOW = "sensorBatteryLow"
    # ---
    MOTION_LIGHT = "lightMotion"
    # ---
    DOORLOCK_OPEN = "doorlockOpened"
    DOORLOCK_CLOSE = "doorlockClosed"
    DOORLOCK_BATTERY_LOW = "doorlockBatteryLow"
    # ---
    DISRUPTED_CONDITIONS = "ringDisruptedConditions"
    # ---
    RECORDING_OFF = "recordingOff"

    @staticmethod
    @cache
    def device_events() -> list[str]:
        return [
            EventType.MOTION.value,
            EventType.NFC_CARD_SCANNED.value,
            EventType.FINGERPRINT_IDENTIFIED.value,
            EventType.RING.value,
            EventType.SMART_DETECT.value,
            EventType.SMART_AUDIO_DETECT.value,
            EventType.SMART_DETECT_LINE.value,
        ]

    @staticmethod
    @cache
    def device_events_set() -> set[str]:
        return set(EventType.device_events())

    @staticmethod
    @cache
    def motion_events() -> list[str]:
        return [EventType.MOTION.value, EventType.SMART_DETECT.value]


@enum.unique
class StateType(str, ValuesEnumMixin, enum.Enum):
    CONNECTED = "CONNECTED"
    CONNECTING = "CONNECTING"
    DISCONNECTED = "DISCONNECTED"


@enum.unique
class ProtectWSPayloadFormat(int, enum.Enum):
    """Websocket Payload formats."""

    JSON = 1
    UTF8String = 2
    NodeBuffer = 3


@enum.unique
class SmartDetectObjectType(str, ValuesEnumMixin, enum.Enum):
    PERSON = "person"
    ANIMAL = "animal"
    VEHICLE = "vehicle"
    LICENSE_PLATE = "licensePlate"
    PACKAGE = "package"
    SMOKE = "alrmSmoke"
    CMONX = "alrmCmonx"
    SIREN = "alrmSiren"
    BABY_CRY = "alrmBabyCry"
    SPEAK = "alrmSpeak"
    BARK = "alrmBark"
    BURGLAR = "alrmBurglar"
    CAR_HORN = "alrmCarHorn"
    GLASS_BREAK = "alrmGlassBreak"
    FACE = "face"
    # old?
    CAR = "car"
    PET = "pet"

    @cached_property
    def audio_type(self) -> SmartDetectAudioType | None:
        return OBJECT_TO_AUDIO_MAP.get(self)


@enum.unique
class SmartDetectAudioType(str, ValuesEnumMixin, enum.Enum):
    SMOKE = "alrmSmoke"
    CMONX = "alrmCmonx"
    SMOKE_CMONX = "smoke_cmonx"
    SIREN = "alrmSiren"
    BABY_CRY = "alrmBabyCry"
    SPEAK = "alrmSpeak"
    BARK = "alrmBark"
    BURGLAR = "alrmBurglar"
    CAR_HORN = "alrmCarHorn"
    GLASS_BREAK = "alrmGlassBreak"


@enum.unique
class DetectionColor(str, ValuesEnumMixin, enum.Enum):
    BLACK = "black"
    BLUE = "blue"
    BROWN = "brown"
    GRAY = "gray"
    GREEN = "green"
    ORANGE = "orange"
    PINK = "pink"
    PURPLE = "purple"
    RED = "red"
    WHITE = "white"
    YELLOW = "yellow"


OBJECT_TO_AUDIO_MAP = {
    SmartDetectObjectType.SMOKE: SmartDetectAudioType.SMOKE,
    SmartDetectObjectType.CMONX: SmartDetectAudioType.CMONX,
    SmartDetectObjectType.SIREN: SmartDetectAudioType.SIREN,
    SmartDetectObjectType.BABY_CRY: SmartDetectAudioType.BABY_CRY,
    SmartDetectObjectType.SPEAK: SmartDetectAudioType.SPEAK,
    SmartDetectObjectType.BARK: SmartDetectAudioType.BARK,
    SmartDetectObjectType.BURGLAR: SmartDetectAudioType.BURGLAR,
    SmartDetectObjectType.CAR_HORN: SmartDetectAudioType.CAR_HORN,
    SmartDetectObjectType.GLASS_BREAK: SmartDetectAudioType.GLASS_BREAK,
}


@enum.unique
class DoorbellMessageType(str, ValuesEnumMixin, enum.Enum):
    LEAVE_PACKAGE_AT_DOOR = "LEAVE_PACKAGE_AT_DOOR"
    DO_NOT_DISTURB = "DO_NOT_DISTURB"
    CUSTOM_MESSAGE = "CUSTOM_MESSAGE"
    IMAGE = "IMAGE"


@enum.unique
class LightModeEnableType(str, ValuesEnumMixin, enum.Enum):
    DARK = "dark"
    ALWAYS = "fulltime"
    NIGHT = "night"


@enum.unique
class LightModeType(str, ValuesEnumMixin, enum.Enum):
    MOTION = "motion"
    WHEN_DARK = "always"
    MANUAL = "off"
    SCHEDULE = "schedule"


@enum.unique
class VideoMode(str, ValuesEnumMixin, enum.Enum):
    DEFAULT = "default"
    HIGH_FPS = "highFps"
    HOMEKIT = "homekit"
    SPORT = "sport"
    SLOW_SHUTTER = "slowShutter"
    # should only be for unadopted devices
    UNKNOWN = "unknown"


@enum.unique
class AudioStyle(str, UnknownValuesEnumMixin, enum.Enum):
    NATURE = "nature"
    NOISE_REDUCED = "noiseReduced"


@enum.unique
class RecordingMode(str, ValuesEnumMixin, enum.Enum):
    ALWAYS = "always"
    NEVER = "never"
    SCHEDULE = "schedule"
    DETECTIONS = "detections"


@enum.unique
class AnalyticsOption(str, ValuesEnumMixin, enum.Enum):
    NONE = "none"
    ANONYMOUS = "anonymous"
    FULL = "full"


@enum.unique
class RecordingType(str, ValuesEnumMixin, enum.Enum):
    TIMELAPSE = "timelapse"
    CONTINUOUS = "rotating"
    DETECTIONS = "detections"


@enum.unique
class ResolutionStorageType(str, ValuesEnumMixin, enum.Enum):
    UHD = "4K"
    HD = "HD"
    FREE = "free"


@enum.unique
class IRLEDMode(str, UnknownValuesEnumMixin, enum.Enum):
    AUTO = "auto"
    ON = "on"
    AUTO_NO_LED = "autoFilterOnly"
    OFF = "off"
    MANUAL = "manual"
    CUSTOM = "custom"
    UNKNOWN = "unknown"


@enum.unique
class MountType(str, ValuesEnumMixin, enum.Enum):
    NONE = "none"
    LEAK = "leak"
    DOOR = "door"
    WINDOW = "window"
    GARAGE = "garage"


@enum.unique
class SensorType(str, ValuesEnumMixin, enum.Enum):
    TEMPERATURE = "temperature"
    LIGHT = "light"
    HUMIDITY = "humidity"


@enum.unique
class SensorStatusType(str, UnknownValuesEnumMixin, enum.Enum):
    OFFLINE = "offline"
    UNKNOWN = "unknown"
    SAFE = "safe"
    NEUTRAL = "neutral"
    LOW = "low"
    HIGH = "high"


@enum.unique
class SleepStateType(str, ValuesEnumMixin, enum.Enum):
    DISCONNECTED = "disconnected"
    AWAKE = "awake"
    START_SLEEP = "goingToSleep"
    ASLEEP = "asleep"
    WAKING = "waking"


@enum.unique
class AutoExposureMode(str, ValuesEnumMixin, enum.Enum):
    MANUAL = "manual"
    AUTO = "auto"
    SHUTTER = "shutter"
    FLICK50 = "flick50"
    FLICK60 = "flick60"


@enum.unique
class FocusMode(str, ValuesEnumMixin, enum.Enum):
    MANUAL = "manual"
    AUTO = "auto"
    ZTRIG = "ztrig"
    TOUCH = "touch"


@enum.unique
class MountPosition(str, UnknownValuesEnumMixin, enum.Enum):
    CEILING = "ceiling"
    WALL = "wall"
    DESK = "desk"
    NONE = "none"
    UNKNOWN = "unknown"


@enum.unique
class GeofencingSetting(str, ValuesEnumMixin, enum.Enum):
    OFF = "off"
    ALL_AWAY = "allAway"


@enum.unique
class MotionAlgorithm(str, ValuesEnumMixin, enum.Enum):
    STABLE = "stable"
    ENHANCED = "enhanced"


@enum.unique
class AudioCodecs(str, ValuesEnumMixin, enum.Enum):
    AAC = "aac"
    VORBIS = "vorbis"
    OPUS = "opus"


@enum.unique
class LowMedHigh(str, ValuesEnumMixin, enum.Enum):
    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"


@enum.unique
class StorageType(str, UnknownValuesEnumMixin, enum.Enum):
    DISK = "hdd"
    RAID = "raid"
    SD_CARD = "sdcard"
    INTERNAL_SSD = "internalSSD"
    UNKNOWN = "UNKNOWN"


@enum.unique
class FirmwareReleaseChannel(str, ValuesEnumMixin, enum.Enum):
    INTERNAL = "internal"
    ALPHA = "alpha"
    BETA = "beta"
    RELEASE_CANDIDATE = "release-candidate"
    RELEASE = "release"


@enum.unique
class ChimeType(int, enum.Enum):
    NONE = 0
    MECHANICAL = 300
    DIGITAL = 1000


@enum.unique
class LockStatusType(str, ValuesEnumMixin, enum.Enum):
    OPEN = "OPEN"
    OPENING = "OPENING"
    CLOSED = "CLOSED"
    CLOSING = "CLOSING"
    JAMMED_WHILE_CLOSING = "JAMMED_WHILE_CLOSING"
    JAMMED_WHILE_OPENING = "JAMMED_WHILE_OPENING"
    FAILED_WHILE_CLOSING = "FAILED_WHILE_CLOSING"
    FAILED_WHILE_OPENING = "FAILED_WHILE_OPENING"
    NOT_CALIBRATED = "NOT_CALIBRATED"
    AUTO_CALIBRATION_IN_PROGRESS = "AUTO_CALIBRATION_IN_PROGRESS"
    CALIBRATION_WAITING_OPEN = "CALIBRATION_WAITING_OPEN"
    CALIBRATION_WAITING_CLOSE = "CALIBRATION_WAITING_CLOSE"


@enum.unique
class PermissionNode(str, UnknownValuesEnumMixin, enum.Enum):
    CREATE = "create"
    READ = "read"
    WRITE = "write"
    DELETE = "delete"
    READ_MEDIA = "readmedia"
    DELETE_MEDIA = "deletemedia"
    READ_LIVE = "readlive"
    UNKNOWN = "unknown"

    @cached_property
    def name(self) -> str:
        """Return the name."""
        return self._name_

    @cached_property
    def value(self) -> str:
        """Return the value."""
        return self._value_


@enum.unique
class HDRMode(str, UnknownValuesEnumMixin, enum.Enum):
    NORMAL = "normal"
    ALWAYS_ON = "superHdr"


@enum.unique
class LensType(str, enum.Enum):
    NONE = "none"
    FULL_360 = "360"
    WIDE = "wide"
    TELESCOPIC = "tele"
    DLSR_17 = "m43"


class DoorbellText(ConstrainedStr):
    max_length = 30


class ICRCustomValue(ConstrainedInt):
    ge = 0
    le = 10


class ICRLuxValue(ConstrainedInt):
    ge = 1
    le = 30


class LEDLevel(ConstrainedInt):
    ge = 0
    le = 6


class PercentInt(ConstrainedInt):
    ge = 0
    le = 100


class TwoByteInt(ConstrainedInt):
    ge = 1
    le = 255


class PercentFloat(ConstrainedFloat):
    ge = 0
    le = 100


class WDRLevel(ConstrainedInt):
    ge = 0
    le = 3


class ICRSensitivity(ConstrainedInt):
    ge = 0
    le = 3


class Percent(ConstrainedFloat):
    ge = 0
    le = 1


class RepeatTimes(ConstrainedInt):
    ge = 1
    le = 6


class PTZPositionDegree(BaseModel):
    pan: float
    tilt: float
    zoom: int


class PTZPositionSteps(BaseModel):
    focus: int
    pan: int
    tilt: int
    zoom: int


class PTZPosition(BaseModel):
    degree: PTZPositionDegree
    steps: PTZPositionSteps


class PTZPresetPosition(BaseModel):
    pan: int
    tilt: int
    zoom: int


class PTZPreset(BaseModel):
    id: str
    name: str
    slot: int
    ptz: PTZPresetPosition


CoordType = Union[Percent, int, float]


# TODO: fix when upgrading to pydantic v2
class Color(BaseColor):
    def __eq__(self, o: object) -> bool:
        if isinstance(o, Color):
            return self.as_hex() == o.as_hex()

        return super().__eq__(o)


class Version(BaseVersion):
    def __str__(self) -> str:
        super_str = super().__str__()
        if self.pre is not None and self.pre[0] == "b":
            super_str = super_str.replace("b", "-beta.")
        return super_str



from __future__ import annotations

from dataclasses import dataclass
from datetime import timedelta
from typing import Optional

import typer

from ..api import ProtectApiClient
from ..cli import base
from ..data import Light

app = typer.Typer(rich_markup_mode="rich")

ARG_DEVICE_ID = typer.Argument(None, help="ID of light to select for subcommands")


@dataclass
class LightContext(base.CliContext):
    devices: dict[str, Light]
    device: Light | None = None


ALL_COMMANDS, DEVICE_COMMANDS = base.init_common_commands(app)


@app.callback(invoke_without_command=True)
def main(ctx: typer.Context, device_id: Optional[str] = ARG_DEVICE_ID) -> None:
    """
    Lights device CLI.

    Returns full list of Viewers without any arguments passed.
    """
    protect: ProtectApiClient = ctx.obj.protect
    context = LightContext(
        protect=ctx.obj.protect,
        device=None,
        devices=protect.bootstrap.lights,
        output_format=ctx.obj.output_format,
    )
    ctx.obj = context

    if device_id is not None and device_id not in ALL_COMMANDS:
        if (device := protect.bootstrap.lights.get(device_id)) is None:
            typer.secho("Invalid light ID", fg="red")
            raise typer.Exit(1)
        ctx.obj.device = device

    if not ctx.invoked_subcommand:
        if device_id in ALL_COMMANDS:
            ctx.invoke(ALL_COMMANDS[device_id], ctx)
            return

        if ctx.obj.device is not None:
            base.print_unifi_obj(ctx.obj.device, ctx.obj.output_format)
            return

        base.print_unifi_dict(ctx.obj.devices)


@app.command()
def camera(ctx: typer.Context, camera_id: Optional[str] = typer.Argument(None)) -> None:
    """Returns or sets tha paired camera for a light."""
    base.require_device_id(ctx)
    obj: Light = ctx.obj.device

    if camera_id is None:
        base.print_unifi_obj(obj.camera, ctx.obj.output_format)
    else:
        protect: ProtectApiClient = ctx.obj.protect
        if (camera_obj := protect.bootstrap.cameras.get(camera_id)) is None:
            typer.secho("Invalid camera ID")
            raise typer.Exit(1)
        base.run(ctx, obj.set_paired_camera(camera_obj))


@app.command()
def set_status_light(ctx: typer.Context, enabled: bool) -> None:
    """Sets status light for light device."""
    base.require_device_id(ctx)
    obj: Light = ctx.obj.device

    base.run(ctx, obj.set_status_light(enabled))


@app.command()
def set_led_level(
    ctx: typer.Context,
    led_level: int = typer.Argument(..., min=1, max=6),
) -> None:
    """Sets brightness of LED on light."""
    base.require_device_id(ctx)
    obj: Light = ctx.obj.device

    base.run(ctx, obj.set_led_level(led_level))


@app.command()
def set_sensitivity(
    ctx: typer.Context,
    sensitivity: int = typer.Argument(..., min=0, max=100),
) -> None:
    """Sets motion sensitivity for the light."""
    base.require_device_id(ctx)
    obj: Light = ctx.obj.device

    base.run(ctx, obj.set_sensitivity(sensitivity))


@app.command()
def set_duration(
    ctx: typer.Context,
    duration: int = typer.Argument(..., min=15, max=900),
) -> None:
    """Sets timeout duration (in seconds) for light."""
    base.require_device_id(ctx)
    obj: Light = ctx.obj.device

    base.run(ctx, obj.set_duration(timedelta(seconds=duration)))



from __future__ import annotations

from collections.abc import Callable, Coroutine, Mapping, Sequence
from dataclasses import dataclass
from enum import Enum
from typing import Any, Optional, TypeVar

import orjson
import typer
from pydantic.v1 import ValidationError

from ..api import ProtectApiClient
from ..data import NVR, ProtectAdoptableDeviceModel, ProtectBaseObject
from ..exceptions import BadRequest, NvrError, StreamError
from ..utils import run_async

T = TypeVar("T")

OPTION_FORCE = typer.Option(False, "-f", "--force", help="Skip confirmation prompt")


class OutputFormatEnum(str, Enum):
    JSON = "json"
    PLAIN = "plain"


@dataclass
class CliContext:
    protect: ProtectApiClient
    output_format: OutputFormatEnum


def run(ctx: typer.Context, func: Coroutine[Any, Any, T]) -> T:
    """Helper method to call async function and clean up API client"""

    async def callback() -> T:
        return_value = await func
        await ctx.obj.protect.close_session()
        return return_value

    try:
        return run_async(callback())
    except (BadRequest, ValidationError, StreamError, NvrError) as err:
        typer.secho(str(err), fg="red")
        raise typer.Exit(1) from err


def json_output(obj: Any) -> None:
    typer.echo(orjson.dumps(obj, option=orjson.OPT_INDENT_2).decode("utf-8"))


def print_unifi_obj(
    obj: ProtectBaseObject | None,
    output_format: OutputFormatEnum,
) -> None:
    """Helper method to print a single protect object"""
    if obj is not None:
        json_output(obj.unifi_dict())
    elif output_format == OutputFormatEnum.JSON:
        json_output(None)


def print_unifi_list(objs: Sequence[ProtectBaseObject]) -> None:
    """Helper method to print a list of protect objects"""
    data = [o.unifi_dict() for o in objs]
    json_output(data)


def print_unifi_dict(objs: Mapping[str, ProtectBaseObject]) -> None:
    """Helper method to print a dictionary of protect objects"""
    data = {k: v.unifi_dict() for k, v in objs.items()}
    json_output(data)


def require_device_id(ctx: typer.Context) -> None:
    """Requires device ID in context"""
    if ctx.obj.device is None:
        typer.secho("Requires a valid device ID to be selected")
        raise typer.Exit(1)


def require_no_device_id(ctx: typer.Context) -> None:
    """Requires no device ID in context"""
    if ctx.obj.device is not None:
        typer.secho("Requires no device ID to be selected")
        raise typer.Exit(1)


def list_ids(ctx: typer.Context) -> None:
    """Requires no device ID. Prints list of "id name" for each device."""
    require_no_device_id(ctx)
    objs: dict[str, ProtectAdoptableDeviceModel] = ctx.obj.devices
    to_print: list[tuple[str, str | None]] = []
    for obj in objs.values():
        name = obj.display_name
        if obj.is_adopted_by_other:
            name = f"{name} [Managed by Another Console]"
        elif obj.is_adopting:
            name = f"{name} [Adopting]"
        elif obj.can_adopt:
            name = f"{name} [Unadopted]"
        elif obj.is_rebooting:
            name = f"{name} [Restarting]"
        elif obj.is_updating:
            name = f"{name} [Updating]"
        elif not obj.is_connected:
            name = f"{name} [Disconnected]"

        to_print.append((obj.id, name))

    if ctx.obj.output_format == OutputFormatEnum.JSON:
        json_output(to_print)
    else:
        for item in to_print:
            typer.echo(f"{item[0]}\t{item[1]}")


def protect_url(ctx: typer.Context) -> None:
    """Gets UniFi Protect management URL."""
    require_device_id(ctx)
    obj: NVR | ProtectAdoptableDeviceModel = ctx.obj.device
    if ctx.obj.output_format == OutputFormatEnum.JSON:
        json_output(obj.protect_url)
    else:
        typer.echo(obj.protect_url)


def is_wired(ctx: typer.Context) -> None:
    """Returns if the device is wired or not."""
    require_device_id(ctx)
    obj: ProtectAdoptableDeviceModel = ctx.obj.device
    json_output(obj.is_wired)


def is_wifi(ctx: typer.Context) -> None:
    """Returns if the device has WiFi or not."""
    require_device_id(ctx)
    obj: ProtectAdoptableDeviceModel = ctx.obj.device
    json_output(obj.is_wifi)


def is_bluetooth(ctx: typer.Context) -> None:
    """Returns if the device has Bluetooth or not."""
    require_device_id(ctx)
    obj: ProtectAdoptableDeviceModel = ctx.obj.device
    json_output(obj.is_bluetooth)


def bridge(ctx: typer.Context) -> None:
    """Returns bridge device if connected via Bluetooth."""
    require_device_id(ctx)
    obj: ProtectAdoptableDeviceModel = ctx.obj.device
    print_unifi_obj(obj.bridge, ctx.obj.output_format)


def set_ssh(ctx: typer.Context, enabled: bool) -> None:
    """
    Sets the isSshEnabled value for device.

    May not have an effect on many device types. Only seems to work for
    Linux and BusyBox based devices (camera, light and viewport).
    """
    require_device_id(ctx)
    obj: ProtectAdoptableDeviceModel = ctx.obj.device
    run(ctx, obj.set_ssh(enabled))


def set_name(ctx: typer.Context, name: Optional[str] = typer.Argument(None)) -> None:
    """Sets name for the device"""
    require_device_id(ctx)
    obj: NVR | ProtectAdoptableDeviceModel = ctx.obj.device
    run(ctx, obj.set_name(name))


def update(ctx: typer.Context, data: str) -> None:
    """
    Updates the device.

    Makes a raw PATCH request to update a device. Advanced usage and usually recommended not to use.
    """
    require_device_id(ctx)
    obj: ProtectAdoptableDeviceModel = ctx.obj.device

    if obj.model is not None:
        run(ctx, obj.api.update_device(obj.model, obj.id, orjson.loads(data)))


def reboot(ctx: typer.Context, force: bool = OPTION_FORCE) -> None:
    """Reboots the device."""
    require_device_id(ctx)
    obj: NVR | ProtectAdoptableDeviceModel = ctx.obj.device

    if force or typer.confirm(f'Confirm reboot of "{obj.name}"" (id: {obj.id})'):
        run(ctx, obj.reboot())


def unadopt(ctx: typer.Context, force: bool = OPTION_FORCE) -> None:
    """Unadopt/Unmanage adopted device."""
    require_device_id(ctx)
    obj: ProtectAdoptableDeviceModel = ctx.obj.device

    if force or typer.confirm(f'Confirm undopt of "{obj.name}"" (id: {obj.id})'):
        run(ctx, obj.unadopt())


def adopt(ctx: typer.Context, name: Optional[str] = typer.Argument(None)) -> None:
    """
    Adopts a device.

    By default, unadopted devices do not show up in the bootstrap. Use
    `uiprotect -u` to show unadopted devices.
    """
    require_device_id(ctx)
    obj: ProtectAdoptableDeviceModel = ctx.obj.device

    run(ctx, obj.adopt(name))


def init_common_commands(
    app: typer.Typer,
) -> tuple[dict[str, Callable[..., Any]], dict[str, Callable[..., Any]]]:
    deviceless_commands: dict[str, Callable[..., Any]] = {}
    device_commands: dict[str, Callable[..., Any]] = {}

    deviceless_commands["list-ids"] = app.command()(list_ids)
    device_commands["protect-url"] = app.command()(protect_url)
    device_commands["is-wired"] = app.command()(is_wired)
    device_commands["is-wifi"] = app.command()(is_wifi)
    device_commands["is-bluetooth"] = app.command()(is_bluetooth)
    device_commands["bridge"] = app.command()(bridge)
    device_commands["set-ssh"] = app.command()(set_ssh)
    device_commands["set-name"] = app.command()(set_name)
    device_commands["update"] = app.command()(update)
    device_commands["reboot"] = app.command()(reboot)
    device_commands["unadopt"] = app.command()(unadopt)
    device_commands["adopt"] = app.command()(adopt)

    return deviceless_commands, device_commands



from __future__ import annotations

import asyncio
import logging
import math
import os
import sys
import time
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, cast

import aiofiles
import aiofiles.os as aos
import av
import dateparser
import typer
from PIL import Image
from rich.progress import (
    BarColumn,
    MofNCompleteColumn,
    Progress,
    TaskProgressColumn,
    TextColumn,
    TimeRemainingColumn,
    track,
)
from sqlalchemy import Column, DateTime, ForeignKey, Integer, String, func, or_, select
from sqlalchemy import event as saevent
from sqlalchemy.ext.asyncio import AsyncEngine, AsyncSession, create_async_engine
from sqlalchemy.orm import Mapped, declarative_base, relationship

from .. import data as d
from ..api import ProtectApiClient
from ..cli import base
from ..utils import (
    format_duration,
    get_local_timezone,
    local_datetime,
    utc_now,
)

if TYPE_CHECKING:
    from click.core import Parameter

app = typer.Typer(rich_markup_mode="rich")
Base = declarative_base()

_LOGGER = logging.getLogger(__name__)


def _on_db_connect(dbapi_con, connection_record) -> None:  # type: ignore[no-untyped-def]
    cursor = dbapi_con.cursor()
    cursor.execute("PRAGMA journal_mode=WAL")
    cursor.execute("PRAGMA synchronous=NORMAL")


@dataclass
class BackupContext(base.CliContext):
    start: datetime
    end: datetime | None
    output: Path
    seperator: str
    thumbnail_format: str
    gif_format: str
    event_format: str
    title_format: str
    max_download: int
    page_size: int
    length_cutoff: timedelta
    _db_engine: AsyncEngine | None = None
    _db_session: AsyncSession | None = None

    @property
    def download_thumbnails(self) -> bool:
        return self.thumbnail_format != ""

    @property
    def download_gifs(self) -> bool:
        return self.gif_format != ""

    @property
    def download_videos(self) -> bool:
        return self.event_format != ""

    @property
    def db_file(self) -> Path:
        return self.output / "events.db"

    @property
    def db_engine(self) -> AsyncEngine:
        if self._db_engine is None:
            self._db_engine = create_async_engine(f"sqlite+aiosqlite:///{self.db_file}")
            self._db_session = None
            saevent.listens_for(self._db_engine.sync_engine, "connect")(_on_db_connect)

        return self._db_engine

    def create_db_session(self) -> AsyncSession:
        return AsyncSession(bind=self.db_engine, expire_on_commit=False)

    async def create_db(self) -> None:
        async with self.db_engine.begin() as conn:
            await conn.run_sync(Base.metadata.create_all)


class EventTypeChoice(str, Enum):
    MOTION = d.EventType.MOTION.value
    RING = d.EventType.RING.value
    SMART_DETECT = d.EventType.SMART_DETECT.value
    SMART_DETECT_LINE = d.EventType.SMART_DETECT_LINE.value


class EventSmartType(Base):  # type: ignore[valid-type,misc]
    __tablename__ = "event_smart_type"

    id = Column(Integer, primary_key=True)
    event_id = Column(String(24), ForeignKey("event.id"), nullable=False)
    smart_type = Column(String(32), index=True)


class Event(Base):  # type: ignore[valid-type,misc]
    __tablename__ = "event"
    __allow_unmapped__ = True

    id = Column(String(24), primary_key=True)
    start_naive = Column(DateTime())
    end_naive = Column(DateTime(), nullable=True)
    camera_mac = Column(String(12), index=True)
    event_type = Column(String(32), index=True)

    smart_detect_types: Mapped[list[EventSmartType]] = relationship(
        "EventSmartType",
        lazy="joined",
        uselist=True,
    )

    _start: datetime | None = None
    _end: datetime | None = None
    _smart_types: set[str] | None = None
    _context: dict[str, str] | None = None
    _glob_context: dict[str, str] | None = None

    @property
    def start(self) -> datetime:
        if self._start is None:
            self._start = self.start_naive.replace(tzinfo=timezone.utc)  # type: ignore[union-attr]
        return self._start

    @property
    def end(self) -> datetime | None:
        if self._end is None and self.end_naive is not None:
            self._end = self.end_naive.replace(tzinfo=timezone.utc)
        return self._end

    @property
    def smart_types(self) -> set[str]:
        if self._smart_types is None:
            self._smart_types = {s.smart_type for s in self.smart_detect_types}  # type: ignore[misc]
        return self._smart_types

    def get_file_context(self, ctx: BackupContext) -> dict[str, str]:
        if self._context is None:
            camera = ctx.protect.bootstrap.get_device_from_mac(self.camera_mac)  # type: ignore[arg-type]
            camera_slug = ""
            display_name = ""
            length = timedelta(seconds=0)
            if camera is not None:
                camera_slug = (
                    camera.display_name.lower().replace(" ", ctx.seperator)
                    + ctx.seperator
                )
                display_name = camera.display_name
            if self.end is not None:
                length = self.end - self.start

            event_type = str(self.event_type)
            event_type_pretty = f"{event_type.title()} Event"
            if event_type in {
                d.EventType.SMART_DETECT.value,
                d.EventType.SMART_DETECT_LINE.value,
            }:
                smart_types = list(self.smart_types)
                smart_types.sort()
                event_type = f"{event_type}[{','.join(smart_types)}]"
                smart_types_title = [s.title() for s in smart_types]
                event_type_pretty = f"Smart Detection ({', '.join(smart_types_title)})"

            start_local = local_datetime(self.start)
            self._context = {
                "year": str(self.start.year),
                "month": str(self.start.month),
                "day": str(self.start.day),
                "hour": str(self.start.hour),
                "minute": str(self.start.minute),
                "datetime": self.start.strftime("%Y-%m-%dT%H-%M-%S%z").replace(
                    "-",
                    ctx.seperator,
                ),
                "date": self.start.strftime("%Y-%m-%d").replace("-", ctx.seperator),
                "time": self.start.strftime("%H-%M-%S%z").replace("-", ctx.seperator),
                "time_sort_pretty": self.start.strftime("%H:%M:%S (%Z)"),
                "time_pretty": self.start.strftime("%I:%M:%S %p (%Z)"),
                "year_local": str(start_local.year),
                "month_local": str(start_local.month),
                "day_local": str(start_local.day),
                "hour_local": str(start_local.hour),
                "minute_local": str(start_local.minute),
                "datetime_local": start_local.strftime("%Y-%m-%dT%H-%M-%S%z").replace(
                    "-",
                    ctx.seperator,
                ),
                "date_local": start_local.strftime("%Y-%m-%d").replace(
                    "-",
                    ctx.seperator,
                ),
                "time_local": start_local.strftime("%H-%M-%S%z").replace(
                    "-",
                    ctx.seperator,
                ),
                "time_sort_pretty_local": start_local.strftime("%H:%M:%S (%Z)"),
                "time_pretty_local": start_local.strftime("%I:%M:%S %p (%Z)"),
                "mac": str(self.camera_mac),
                "camera_name": display_name,
                "camera_slug": camera_slug,
                "event_type": event_type,
                "event_type_pretty": event_type_pretty,
                "length_pretty": format_duration(length),
                "sep": ctx.seperator,
            }

            self._context["title"] = ctx.title_format.format(**self._context)
        return self._context

    def get_glob_file_context(self, ctx: BackupContext) -> dict[str, str]:
        if self._glob_context is None:
            self._glob_context = self.get_file_context(ctx).copy()
            self._glob_context["camera_slug"] = "*"
            self._glob_context["camera_name"] = "*"
        return self._glob_context

    def get_thumbnail_path(self, ctx: BackupContext) -> Path:
        context = self.get_file_context(ctx)
        file_path = ctx.thumbnail_format.format(**context)
        return ctx.output / file_path

    def get_existing_thumbnail_path(self, ctx: BackupContext) -> Path | None:
        context = self.get_glob_file_context(ctx)
        file_path = ctx.thumbnail_format.format(**context)

        paths = list(ctx.output.glob(file_path))
        if paths:
            return paths[0]
        return None

    def get_gif_path(self, ctx: BackupContext) -> Path:
        context = self.get_file_context(ctx)
        file_path = ctx.gif_format.format(**context)
        return ctx.output / file_path

    def get_existing_gif_path(self, ctx: BackupContext) -> Path | None:
        context = self.get_glob_file_context(ctx)
        file_path = ctx.gif_format.format(**context)

        paths = list(ctx.output.glob(file_path))
        if paths:
            return paths[0]
        return None

    def get_event_path(self, ctx: BackupContext) -> Path:
        context = self.get_file_context(ctx)
        file_path = ctx.event_format.format(**context)
        return ctx.output / file_path

    def get_existing_event_path(self, ctx: BackupContext) -> Path | None:
        context = self.get_glob_file_context(ctx)
        file_path = ctx.event_format.format(**context)

        paths = list(ctx.output.glob(file_path))
        if paths:
            return paths[0]
        return None


@dataclass
class QueuedDownload:
    task: asyncio.Task[bool] | None
    args: list[Any]


def relative_datetime(ctx: typer.Context, value: str, param: Parameter) -> datetime:
    if dt := dateparser.parse(value):
        return dt

    raise typer.BadParameter(
        "Must be a ISO 8601 format or human readable relative format",
        ctx,
        param,
    )


_DownloadEventQueue = asyncio.Queue[QueuedDownload]

OPTION_OUTPUT = typer.Option(
    None,
    help="Base dir for creating files. Defaults to $PWD.",
    envvar="UFP_BACKUP_OUTPUT",
)
OPTION_START = typer.Option(
    None,
    "-s",
    "--start",
    help="Cutoff for start of backup. Defaults to start of recording for NVR.",
    envvar="UFP_BACKUP_START",
)
OPTION_PAGE_SIZE = typer.Option(
    1000,
    "--page-size",
    help="Number of events fetched at once from local database. Increases memory usage.",
)
OPTION_LENGTH_CUTOFF = typer.Option(
    timedelta(hours=1).total_seconds(),
    "--length-cutoff",
    help="Event size cutoff for detecting abnormal events (in seconds).",
)
OPTION_END = typer.Option(
    None,
    "-e",
    "--end",
    help="Cutoff for end of backup. Defaults to now.",
    envvar="UFP_BACKUP_END",
)
OPTION_EVENT_TYPES = typer.Option(
    list(EventTypeChoice),
    "-t",
    "--event-type",
    help="Events to export. Can be used multiple time.",
)
OPTION_SMART_TYPES = typer.Option(
    list(d.SmartDetectObjectType),
    "-m",
    "--smart-type",
    help="Smart Detection types to export. Can be used multiple time.",
)
OPTION_SPERATOR = typer.Option("-", "--sep", help="Separator used for formatting.")
OPTION_THUMBNAIL_FORMAT = typer.Option(
    "{year}/{month}/{day}/{hour}/{datetime}{sep}{mac}{sep}{camera_slug}{event_type}{sep}thumb.jpg",
    "--thumb-format",
    help='Filename format to save event thumbnails to. Set to empty string ("") to skip saving event thumbnails.',
)
OPTION_GIF_FORMAT = typer.Option(
    "{year}/{month}/{day}/{hour}/{datetime}{sep}{mac}{sep}{camera_slug}{event_type}{sep}animated.gif",
    "--gif-format",
    help='Filename format to save event gifs to. Set to empty string ("") to skip saving event gif.',
)
OPTION_EVENT_FORMAT = typer.Option(
    "{year}/{month}/{day}/{hour}/{datetime}{sep}{mac}{sep}{camera_slug}{event_type}.mp4",
    "--event-format",
    help='Filename format to save event gifs to. Set to empty string ("") to skip saving event videos.',
)
OPTION_TITLE_FORMAT = typer.Option(
    "{time_sort_pretty_local} {sep} {camera_name} {sep} {event_type_pretty} {sep} {length_pretty}",
    "--title-format",
    help="Format to use to tag title for video metadata.",
)
OPTION_VERBOSE = typer.Option(False, "-v", "--verbose", help="Debug logging.")
OPTION_MAX_DOWNLOAD = typer.Option(
    5,
    "-d",
    "--max-download",
    help="Max number of concurrent downloads. Adds additional loads to NVR.",
)


def _setup_logger(verbose: bool) -> None:
    console_handler = logging.StreamHandler()
    log_format = "[%(asctime)s] %(levelname)s - %(message)s"
    if verbose:
        console_handler.setLevel(logging.DEBUG)
    elif sys.stdout.isatty():
        console_handler.setLevel(logging.WARNING)
        log_format = "%(message)s"
    else:
        console_handler.setLevel(logging.INFO)

    formatter = logging.Formatter(log_format)
    console_handler.setFormatter(formatter)

    root_logger = logging.getLogger("uiprotect")
    root_logger.setLevel(logging.DEBUG)
    root_logger.addHandler(console_handler)


@app.callback()
def main(
    ctx: typer.Context,
    start: Optional[str] = OPTION_START,
    end: Optional[str] = OPTION_END,
    output_folder: Optional[Path] = OPTION_OUTPUT,
    thumbnail_format: str = OPTION_THUMBNAIL_FORMAT,
    gif_format: str = OPTION_GIF_FORMAT,
    event_format: str = OPTION_EVENT_FORMAT,
    title_format: str = OPTION_TITLE_FORMAT,
    verbose: bool = OPTION_VERBOSE,
    max_download: int = OPTION_MAX_DOWNLOAD,
    page_size: int = OPTION_PAGE_SIZE,
    length_cutoff: int = OPTION_LENGTH_CUTOFF,
    seperator: str = OPTION_SPERATOR,
) -> None:
    """
    Backup CLI.

    The backup CLI is still very WIP in progress and consider experimental and potentially unstable (interface may change in the future).
    """
    _setup_logger(verbose)

    protect: ProtectApiClient = ctx.obj.protect
    local_tz = get_local_timezone()

    if start is None:
        start_dt = protect.bootstrap.recording_start
    else:
        start_dt = relative_datetime(ctx, start, ctx.command.params[0])
        start_dt = start_dt.replace(tzinfo=local_tz)
    if start_dt is None:
        start_dt = utc_now()

    end_dt = None
    if end is not None:
        end_dt = relative_datetime(ctx, end, ctx.command.params[1])
        end_dt = end_dt.replace(tzinfo=local_tz)

    if output_folder is None:
        output_folder = Path(os.getcwd())

    context = BackupContext(
        protect=ctx.obj.protect,
        start=start_dt,
        end=end_dt,
        output_format=ctx.obj.output_format,
        output=output_folder,
        thumbnail_format=thumbnail_format,
        gif_format=gif_format,
        event_format=event_format,
        title_format=title_format,
        max_download=max_download,
        page_size=page_size,
        length_cutoff=timedelta(seconds=length_cutoff),
        seperator=seperator,
    )
    ctx.obj = context


def _wipe_files(ctx: BackupContext, no_input: bool) -> None:
    if not no_input and not typer.confirm(
        "Are you sure you want to delete all existing thumbnails and video clips?",
    ):
        raise typer.Exit(1)

    if ctx.db_file.exists():
        os.remove(ctx.db_file)

    for path in track(ctx.output.glob("**/*.jpg"), description="Deleting Thumbnails"):
        os.remove(path)

    for path in track(ctx.output.glob("**/*.mp4"), description="Deleting Clips"):
        os.remove(path)


async def _newest_event(ctx: BackupContext) -> Event | None:
    db = ctx.create_db_session()
    async with db:
        result = await db.execute(select(Event).order_by(Event.start_naive.desc()))
        return result.scalars().first()


async def _prune_events(ctx: BackupContext) -> int:
    _LOGGER.debug("Pruning events before %s", ctx.start)

    deleted = 0
    db = ctx.create_db_session()
    async with db:
        result = await db.execute(
            select(Event).join(EventSmartType).where(Event.start_naive < ctx.start),
        )
        for event in track(result.unique().scalars(), description="Pruning Events"):
            thumb_path = event.get_thumbnail_path(ctx)
            if thumb_path.exists():
                _LOGGER.debug("Delete file %s", thumb_path)
                await aos.remove(thumb_path)

            event_path = event.get_event_path(ctx)
            if event_path.exists():
                _LOGGER.debug("Delete file %s", event_path)
                await aos.remove(event_path)

            if event.event_type in {
                d.EventType.SMART_DETECT.value,
                d.EventType.SMART_DETECT_LINE.value,
            }:
                for smart_type in event.smart_detect_types:
                    await db.delete(smart_type)
            await db.delete(event)
            deleted += 1
        await db.commit()

    return deleted


async def _update_event(ctx: BackupContext, event: d.Event) -> None:
    if event.camera is None:
        return

    db = ctx.create_db_session()
    to_delete: list[EventSmartType] = []
    async with db:
        result = await db.execute(select(Event).where(Event.id == event.id))
        db_event = result.scalars().first()
        do_insert = False
        if db_event is None:
            db_event = Event(id=event.id)
            do_insert = True
        db_event.start_naive = event.start
        db_event.end_naive = event.end
        db_event.camera_mac = event.camera.mac
        db_event.event_type = event.type.value

        if event.type in {
            d.EventType.SMART_DETECT.value,
            d.EventType.SMART_DETECT_LINE.value,
        }:
            types = {e.value for e in event.smart_detect_types}

            result = await db.execute(
                select(EventSmartType).where(EventSmartType.event_id == event.id),
            )
            for event_smart_type in result.unique().scalars():
                event_type = cast(EventSmartType, event_smart_type)
                if event_type.smart_type not in types:
                    to_delete.append(event_type)
                else:
                    types.remove(event_type.smart_type)

            for smart_type_str in types:
                db.add(EventSmartType(event_id=event.id, smart_type=smart_type_str))

        if do_insert:
            db.add(db_event)
        for smart_type in to_delete:
            await db.delete(smart_type)
        await db.commit()


async def _update_ongoing_events(ctx: BackupContext) -> int:
    db = ctx.create_db_session()
    async with db:
        result = await db.execute(
            select(Event)
            .where(Event.event_type != "ring")
            .where(Event.end_naive is None),  # type: ignore[arg-type]
        )

        events = list(result.unique().scalars())

    if len(events) == 0:
        return 0
    for event in track(events, description="Updating Events"):
        event_id = cast(str, event.id)
        await _update_event(ctx, await ctx.protect.get_event(event_id))
    return len(events)


async def _update_events(ctx: BackupContext) -> int:
    # update any events that are still set as ongoing in the database
    updated_ongoing = await _update_ongoing_events(ctx)
    start = ctx.start
    end = ctx.end or utc_now()
    processed: set[str] = set()

    total = int((end - ctx.start).total_seconds())
    _LOGGER.debug("total: %s: %s %s", total, start, end)

    prev_start = start
    with Progress() as pb:
        task_id = pb.add_task("Fetching New Events", total=total)
        task = pb.tasks[0]
        pb.refresh()
        while not pb.finished:
            progress = int((start - prev_start).total_seconds())
            pb.update(task_id, advance=progress)
            _LOGGER.debug(
                "progress: +%s: %s/%s: %s %s",
                progress,
                task.completed,
                task.total,
                start,
                end,
            )

            events = await ctx.protect.get_events(
                start,
                end,
                limit=100,
                types=[
                    d.EventType.MOTION,
                    d.EventType.RING,
                    d.EventType.SMART_DETECT,
                    d.EventType.SMART_DETECT_LINE,
                ],
            )

            prev_start = start
            count = 0
            for event in events:
                start = event.start
                if event.id not in processed:
                    count += 1
                    processed.add(event.id)
                    await _update_event(ctx, event)

            if start == prev_start and count == 0:
                pb.update(task_id, completed=total)

    return updated_ongoing + len(processed)


async def _download_watcher(
    count: int,
    tasks: _DownloadEventQueue,
    no_error_flag: asyncio.Event,
) -> int:
    processed = 0
    loop = asyncio.get_running_loop()
    downloaded = 0
    last_print = time.monotonic()
    while processed < count:
        download = await tasks.get()
        task = download.task
        if task is None:
            processed += 1
            continue

        retries = 0
        while True:
            try:
                await task
            except asyncio.CancelledError:
                return downloaded
            except Exception:
                pass

            event: Event = download.args[1]
            if exception := task.exception():
                no_error_flag.clear()
                if retries < 5:
                    wait = math.pow(2, retries)
                    _LOGGER.warning(
                        "Exception while downloading event (%s): %s. Retring in %s second(s)",
                        event.id,
                        exception,
                        wait,
                    )
                    await asyncio.sleep(wait)
                    retries += 1
                    task = loop.create_task(_download_event(*download.args))
                else:
                    _LOGGER.error("Failed to download event %s", event.id)

            if exception is None or retries >= 5:
                no_error_flag.set()
                processed += 1
                now = time.monotonic()
                if now - last_print > 60:
                    _LOGGER.info(
                        "Processed %s/%s (%.2f%%) events",
                        processed,
                        count,
                        processed / count,
                    )
                    last_print = now
                if exception is None and task.result():
                    downloaded += 1
                break
    return downloaded


def _verify_thumbnail(path: Path) -> bool:
    try:
        image = Image.open(path)
        image.verify()
    # no docs on what exception could be
    except Exception:
        return False
    return True


async def _download_event_thumb(
    ctx: BackupContext,
    event: Event,
    verify: bool,
    force: bool,
    animated: bool = False,
) -> bool:
    if animated:
        thumb_type = "gif"
        thumb_path = event.get_gif_path(ctx)
        existing_thumb_path = event.get_existing_gif_path(ctx)
    else:
        thumb_type = "thumbnail"
        thumb_path = event.get_thumbnail_path(ctx)
        existing_thumb_path = event.get_existing_thumbnail_path(ctx)

    if force and existing_thumb_path:
        _LOGGER.debug("Delete file %s", existing_thumb_path)
        await aos.remove(existing_thumb_path)

    if existing_thumb_path and str(existing_thumb_path) != str(thumb_path):
        _LOGGER.debug(
            "Rename event %s file %s: %s %s %s: %s",
            thumb_type,
            event.id,
            event.start,
            event.end,
            event.event_type,
            thumb_path,
        )
        await aos.makedirs(thumb_path.parent, exist_ok=True)
        await aos.rename(existing_thumb_path, thumb_path)

    if (
        verify
        and thumb_path.exists()
        and not await asyncio.get_running_loop().run_in_executor(
            None, _verify_thumbnail, thumb_path
        )
    ):
        _LOGGER.warning(
            "Corrupted event %s file for event (%s), redownloading",
            thumb_type,
            event.id,
        )
        await aos.remove(thumb_path)

    if not thumb_path.exists():
        _LOGGER.debug(
            "Download event %s %s: %s %s: %s",
            thumb_type,
            event.id,
            event.start,
            event.event_type,
            thumb_path,
        )
        event_id = str(event.id)
        if animated:
            thumbnail = await ctx.protect.get_event_animated_thumbnail(event_id)
        else:
            thumbnail = await ctx.protect.get_event_thumbnail(event_id)
        if thumbnail is not None:
            await aos.makedirs(thumb_path.parent, exist_ok=True)
            async with aiofiles.open(thumb_path, mode="wb") as f:
                await f.write(thumbnail)
            return True
    return False


def _verify_video_file(  # type: ignore[return]
    path: Path,
    length: float,
    width: int,
    height: int,
    title: str,
) -> tuple[bool, bool]:
    try:
        with av.open(str(path)) as video:
            slength = float(
                video.streams.video[0].duration * video.streams.video[0].time_base,  # type: ignore[operator]
            )
            valid = (
                (slength / length) > 0.80  # export is fuzzy
                and video.streams.video[0].codec_context.width == width
                and video.streams.video[0].codec_context.height == height
            )
            metadata_valid = False
            if valid:
                metadata_valid = bool(video.metadata["title"] == title)
            return valid, metadata_valid

    # no docs on what exception could be
    except Exception:
        return False, False


def _add_metadata(path: Path, creation: datetime, title: str) -> bool:
    creation = local_datetime(creation)
    output_path = path.parent / path.name.replace(".mp4", ".metadata.mp4")

    success = True
    try:
        with (
            av.open(str(path)) as input_file,
            av.open(str(output_path), "w") as output_file,
        ):
            for key, value in input_file.metadata.items():
                output_file.metadata[key] = value
            output_file.metadata["creation_time"] = creation.isoformat()
            output_file.metadata["title"] = title
            output_file.metadata["year"] = creation.date().isoformat()
            output_file.metadata["release"] = creation.date().isoformat()

            in_to_out: dict[str, Any] = {}
            for stream in input_file.streams:
                in_to_out[stream] = output_file.add_stream(template=stream)  # type: ignore[index]
                in_to_out[stream].metadata["creation_time"] = creation.isoformat()  # type: ignore[index]

            for packet in input_file.demux(list(in_to_out)):
                if packet.dts is None:
                    continue

                packet.stream = in_to_out[packet.stream]  # type: ignore[index]
                try:
                    output_file.mux(packet)  # type: ignore[arg-type]
                # some frames may be corrupted on disk from NVR
                except ValueError:
                    continue
    # no docs on what exception could be
    except Exception:
        success = False
    finally:
        if success:
            os.remove(path)
            output_path.rename(path)
        elif output_path.exists():
            os.remove(output_path)
    return success


async def _download_event_video(
    ctx: BackupContext,
    camera: d.Camera,
    event: Event,
    verify: bool,
    force: bool,
) -> bool:
    event_path = event.get_event_path(ctx)
    existing_event_path = event.get_existing_event_path(ctx)
    if force and existing_event_path:
        _LOGGER.debug("Delete file %s", existing_event_path)
        await aos.remove(existing_event_path)

    if existing_event_path and str(existing_event_path) != str(event_path):
        _LOGGER.debug(
            "Rename event file %s: %s %s %s: %s",
            event.id,
            event.start,
            event.end,
            event.event_type,
            event_path,
        )
        await aos.makedirs(event_path.parent, exist_ok=True)
        await aos.rename(existing_event_path, event_path)

    metadata_valid = True
    if verify and event_path.exists():
        valid = False
        if event.end is not None:
            valid, metadata_valid = await asyncio.get_running_loop().run_in_executor(
                None,
                _verify_video_file,
                event_path,
                (event.end - event.start).total_seconds(),
                camera.channels[0].width,
                camera.channels[0].height,
                event.get_file_context(ctx)["title"],
            )

        if not valid:
            _LOGGER.warning(
                "Corrupted video file for event (%s), redownloading",
                event.id,
            )
            await aos.remove(event_path)

    downloaded = False
    if not event_path.exists() and event.end is not None:
        _LOGGER.debug(
            "Download event %s: %s %s %s: %s",
            event.id,
            event.start,
            event.end,
            event.event_type,
            event_path,
        )
        await aos.makedirs(event_path.parent, exist_ok=True)
        await camera.get_video(event.start, event.end, output_file=event_path)
        downloaded = True

    if (downloaded or not metadata_valid) and event.end is not None:
        file_context = event.get_file_context(ctx)
        if not await asyncio.get_running_loop().run_in_executor(
            None, _add_metadata, event_path, event.start, file_context["title"]
        ):
            _LOGGER.warning("Failed to write metadata for event (%s)", event.id)
    return downloaded


async def _download_event(
    ctx: BackupContext,
    event: Event,
    verify: bool,
    force: bool,
    pb: Progress,
) -> bool:
    downloaded = False
    camera = ctx.protect.bootstrap.get_device_from_mac(event.camera_mac)  # type: ignore[arg-type]
    if camera is not None:
        camera = cast(d.Camera, camera)
        downloads = []
        if ctx.download_thumbnails:
            downloads.append(_download_event_thumb(ctx, event, verify, force))
        if ctx.download_gifs:
            downloads.append(
                _download_event_thumb(ctx, event, verify, force, animated=True),
            )
        if ctx.download_thumbnails:
            downloads.append(_download_event_video(ctx, camera, event, verify, force))

        downloaded = any(await asyncio.gather(*downloads))
    pb.update(pb.tasks[0].id, advance=1)
    return downloaded


# TODO
async def _download_events(
    ctx: BackupContext,
    event_types: list[d.EventType],
    smart_types: list[d.SmartDetectObjectType],
    verify: bool,
    force: bool,
) -> tuple[int, int]:
    start = ctx.start
    end = ctx.end or utc_now()
    db = ctx.create_db_session()
    async with db:
        count_query = (
            select(func.count(Event.id))
            .where(Event.event_type.in_([e.value for e in event_types]))
            .where(Event.start_naive >= start)
            .where(or_(Event.end_naive <= end, Event.end_naive is None))  # type: ignore[arg-type]
        )
        count = cast(int, (await db.execute(count_query)).scalar())
        _LOGGER.info("Downloading %s events", count)

        columns = [
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            TaskProgressColumn(),
            MofNCompleteColumn(),
            TimeRemainingColumn(),
        ]
        with Progress(*columns) as pb:
            task_id = pb.add_task("Downloading Events", total=count)

            query = (
                select(Event)
                .where(Event.event_type.in_([e.value for e in event_types]))
                .where(Event.start_naive >= start)
                .where(or_(Event.end_naive <= end, Event.end_naive is None))  # type: ignore[arg-type]
                .limit(ctx.page_size)
            )
            smart_types_set = {s.value for s in smart_types}
            loop = asyncio.get_running_loop()
            tasks: _DownloadEventQueue = asyncio.Queue(maxsize=ctx.max_download - 1)
            no_error_flag = asyncio.Event()
            no_error_flag.set()
            watcher_task = loop.create_task(
                _download_watcher(count, tasks, no_error_flag),
            )

            offset = 0
            page = query
            while offset < count:
                result = await db.execute(page)
                for event in result.unique().scalars():
                    if event.end is None:
                        continue

                    length = event.end - event.start
                    if length > ctx.length_cutoff:
                        _LOGGER.warning(
                            "Skipping event %s because it is too long (%s)",
                            event.id,
                            length,
                        )
                        await tasks.put(QueuedDownload(task=None, args=[]))
                        continue
                    # ensure no tasks are currently in a retry state
                    await no_error_flag.wait()

                    if event.event_type in {
                        d.EventType.SMART_DETECT.value,
                        d.EventType.SMART_DETECT_LINE.value,
                    } and not event.smart_types.intersection(smart_types_set):
                        continue

                    task = loop.create_task(
                        _download_event(ctx, event, verify, force, pb),
                    )
                    # waits for a free processing slot
                    await tasks.put(
                        QueuedDownload(task=task, args=[ctx, event, verify, force, pb]),
                    )

                offset += ctx.page_size
                page = query.offset(offset)

            try:
                await watcher_task
                downloaded = watcher_task.result()
            except asyncio.CancelledError:
                downloaded = 0
            pb.update(task_id, completed=count)
    return count, downloaded


async def _events(
    ctx: BackupContext,
    event_types: list[d.EventType],
    smart_types: list[d.SmartDetectObjectType],
    prune: bool,
    force: bool,
    verify: bool,
    no_input: bool,
) -> None:
    try:
        await ctx.create_db()

        if prune and not force:
            _LOGGER.warning("Pruned %s old event(s)", await _prune_events(ctx))

        original_start = ctx.start
        if not force:
            event = await _newest_event(ctx)
            if event is not None:
                ctx.start = event.start

        _LOGGER.warning("Updated %s event(s)", await _update_events(ctx))
        ctx.start = original_start
        count, downloaded = await _download_events(
            ctx,
            event_types,
            smart_types,
            verify,
            force,
        )
        verified = count - downloaded
        _LOGGER.warning(
            "Total events: %s. Verified %s existing event(s). Downloaded %s new event(s)",
            count,
            verified,
            downloaded,
        )
    finally:
        _LOGGER.debug("Cleaning up Protect connection/database...")
        await ctx.protect.close_session()
        await ctx.db_engine.dispose()


@app.command(name="events")
def events_cmd(
    ctx: typer.Context,
    event_types: list[EventTypeChoice] = OPTION_EVENT_TYPES,
    smart_types: list[d.SmartDetectObjectType] = OPTION_SMART_TYPES,
    prune: bool = typer.Option(
        False,
        "-p",
        "--prune",
        help="Prune events older then start.",
    ),
    force: bool = typer.Option(
        False,
        "-f",
        "--force",
        help="Force update all events and redownload all clips.",
    ),
    verify: bool = typer.Option(
        False,
        "-v",
        "--verify",
        help="Verifies files on disk.",
    ),
    no_input: bool = typer.Option(False, "--no-input"),
) -> None:
    """Backup thumbnails and video clips for camera events."""
    # surpress av logging messages
    av.logging.set_level(av.logging.PANIC)
    ufp_events = [d.EventType(e.value) for e in event_types]
    if prune and force:
        _wipe_files(ctx.obj, no_input)
    asyncio.run(
        _events(ctx.obj, ufp_events, smart_types, prune, force, verify, no_input),
    )



"""UniFi Protect User models."""

from __future__ import annotations

from datetime import datetime
from functools import cache
from typing import Any

from pydantic.v1.fields import PrivateAttr

from .base import ProtectBaseObject, ProtectModel, ProtectModelWithId
from .types import ModelType, PermissionNode


class Permission(ProtectBaseObject):
    raw_permission: str
    model: ModelType
    nodes: set[PermissionNode]
    obj_ids: set[str] | None

    @classmethod
    def unifi_dict_to_dict(cls, data: dict[str, Any]) -> dict[str, Any]:
        permission = data.get("rawPermission", "")
        parts = permission.split(":")
        if len(parts) < 2:
            raise ValueError(f"Invalid permission: {permission}")

        data["model"] = ModelType(parts[0])
        if parts[1] == "*":
            data["nodes"] = list(PermissionNode)
        else:
            data["nodes"] = [PermissionNode(n) for n in parts[1].split(",")]

        if len(parts) == 3 and parts[2] != "*":
            if parts[2] == "$":
                data["obj_ids"] = ["self"]
            else:
                data["obj_ids"] = parts[2].split(",")

        return super().unifi_dict_to_dict(data)

    def unifi_dict(  # type: ignore[override]
        self,
        data: dict[str, Any] | None = None,
        exclude: set[str] | None = None,
    ) -> str:
        return self.raw_permission

    @property
    def objs(self) -> list[ProtectModelWithId] | None:
        if self.obj_ids == {"self"} or self.obj_ids is None:
            return None

        devices = getattr(self._api.bootstrap, self.model.devices_key)
        return [devices[oid] for oid in self.obj_ids]


class Group(ProtectModelWithId):
    name: str
    permissions: list[Permission]
    type: str
    is_default: bool

    @classmethod
    def unifi_dict_to_dict(cls, data: dict[str, Any]) -> dict[str, Any]:
        if "permissions" in data:
            permissions = data.pop("permissions")
            data["permissions"] = [{"rawPermission": p} for p in permissions]

        return super().unifi_dict_to_dict(data)


class UserLocation(ProtectModel):
    is_away: bool
    latitude: float | None
    longitude: float | None


class CloudAccount(ProtectModelWithId):
    first_name: str
    last_name: str
    email: str
    user_id: str
    name: str
    location: UserLocation | None
    profile_img: str | None = None

    @classmethod
    @cache
    def _get_unifi_remaps(cls) -> dict[str, str]:
        return {**super()._get_unifi_remaps(), "user": "userId"}

    def unifi_dict(
        self,
        data: dict[str, Any] | None = None,
        exclude: set[str] | None = None,
    ) -> dict[str, Any]:
        data = super().unifi_dict(data=data, exclude=exclude)

        # id and cloud ID are always the same
        if "id" in data:
            data["cloudId"] = data["id"]
        if "location" in data and data["location"] is None:
            del data["location"]

        return data

    @property
    def user(self) -> User:
        return self._api.bootstrap.users[self.user_id]


class UserFeatureFlags(ProtectBaseObject):
    notifications_v2: bool


class User(ProtectModelWithId):
    permissions: list[Permission]
    last_login_ip: str | None
    last_login_time: datetime | None
    is_owner: bool
    enable_notifications: bool
    has_accepted_invite: bool
    all_permissions: list[Permission]
    scopes: list[str] | None = None
    location: UserLocation | None
    name: str
    first_name: str
    last_name: str
    email: str | None
    local_username: str
    group_ids: list[str]
    cloud_account: CloudAccount | None
    feature_flags: UserFeatureFlags

    # TODO:
    # settings
    # alertRules
    # notificationsV2
    # notifications
    # cloudProviders

    _groups: list[Group] | None = PrivateAttr(None)
    _perm_cache: dict[str, bool] = PrivateAttr({})

    def __init__(self, **data: Any) -> None:
        if "permissions" in data:
            permissions = data.pop("permissions")
            data["permissions"] = [
                {"raw_permission": p} if isinstance(p, str) else p for p in permissions
            ]
        if "allPermissions" in data:
            permissions = data.pop("allPermissions")
            data["allPermissions"] = [
                {"raw_permission": p} if isinstance(p, str) else p for p in permissions
            ]

        super().__init__(**data)

    @classmethod
    def unifi_dict_to_dict(cls, data: dict[str, Any]) -> dict[str, Any]:
        if "permissions" in data:
            permissions = data.pop("permissions")
            data["permissions"] = [{"rawPermission": p} for p in permissions]
        if "allPermissions" in data:
            permissions = data.pop("allPermissions")
            data["allPermissions"] = [{"rawPermission": p} for p in permissions]

        return super().unifi_dict_to_dict(data)

    @classmethod
    @cache
    def _get_unifi_remaps(cls) -> dict[str, str]:
        return {**super()._get_unifi_remaps(), "groups": "groupIds"}

    def unifi_dict(
        self,
        data: dict[str, Any] | None = None,
        exclude: set[str] | None = None,
    ) -> dict[str, Any]:
        data = super().unifi_dict(data=data, exclude=exclude)

        if "location" in data and data["location"] is None:
            del data["location"]

        return data

    @property
    def groups(self) -> list[Group]:
        """
        Groups the user is in

        Will always be empty if the user only has read only access.
        """
        if self._groups is not None:
            return self._groups

        self._groups = [
            self._api.bootstrap.groups[g]
            for g in self.group_ids
            if g in self._api.bootstrap.groups
        ]
        return self._groups

    def can(
        self,
        model: ModelType,
        node: PermissionNode,
        obj: ProtectModelWithId | None = None,
    ) -> bool:
        """Checks if a user can do a specific action"""
        check_self = False
        if model is self.model and obj is not None and obj.id == self.id:
            perm_str = f"{model.value}:{node.value}:$"
            check_self = True
        else:
            perm_str = (
                f"{model.value}:{node.value}:{obj.id if obj is not None else '*'}"
            )
        if perm_str in self._perm_cache:
            return self._perm_cache[perm_str]

        for perm in self.all_permissions:
            if model is not perm.model or node not in perm.nodes:
                continue
            if perm.obj_ids is None:
                self._perm_cache[perm_str] = True
                return True
            if check_self and perm.obj_ids == {"self"}:
                self._perm_cache[perm_str] = True
                return True
            if perm.obj_ids is not None and obj is not None and obj.id in perm.obj_ids:
                self._perm_cache[perm_str] = True
                return True
        self._perm_cache[perm_str] = False
        return False



from __future__ import annotations

import asyncio
import contextlib
import json
import logging
import math
import os
import re
import socket
import sys
import time
import zoneinfo
from collections import Counter
from collections.abc import Callable, Coroutine, Iterable
from copy import deepcopy
from datetime import datetime, timedelta, timezone, tzinfo
from decimal import Decimal
from enum import Enum
from functools import cache, lru_cache, partial
from hashlib import sha224
from http.cookies import Morsel
from inspect import isclass
from ipaddress import IPv4Address, IPv6Address, ip_address
from operator import attrgetter
from pathlib import Path
from typing import TYPE_CHECKING, Any, TypeVar, Union, overload
from uuid import UUID

import jwt
from aiohttp import ClientResponse
from pydantic.v1.fields import SHAPE_DICT, SHAPE_LIST, SHAPE_SET, ModelField
from pydantic.v1.utils import to_camel

from .data.types import (
    Color,
    SmartDetectAudioType,
    SmartDetectObjectType,
    Version,
    VideoMode,
)
from .exceptions import NvrError

if TYPE_CHECKING:
    from uiprotect.api import ProtectApiClient
    from uiprotect.data import CoordType, Event
    from uiprotect.data.bootstrap import WSStat

if sys.version_info[:2] < (3, 11):
    from async_timeout import timeout as asyncio_timeout
else:
    from asyncio import timeout as asyncio_timeout  # noqa: F401

T = TypeVar("T")

DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S"
DEBUG_ENV = "UFP_DEBUG"
PROGRESS_CALLABLE = Callable[[int, str], Coroutine[Any, Any, None]]
SNAKE_CASE_KEYS = [
    "life_span",
    "bad_sector",
    "total_bytes",
    "used_bytes",
    "space_type",
]
TIMEZONE_GLOBAL: tzinfo | None = None

SNAKE_CASE_MATCH_1 = re.compile("(.)([A-Z0-9][a-z]+)")
SNAKE_CASE_MATCH_2 = re.compile("__([A-Z0-9])")
SNAKE_CASE_MATCH_3 = re.compile("([a-z0-9])([A-Z])")

_LOGGER = logging.getLogger(__name__)

RELEASE_CACHE = Path(__file__).parent / "release_cache.json"

_CREATE_TYPES = {IPv6Address, IPv4Address, UUID, Color, Decimal, Path, Version}
_BAD_UUID = "00000000-0000-00 0- 000-000000000000"

IP_TYPES = {
    Union[IPv4Address, str, None],
    Union[IPv4Address, str],
    Union[IPv6Address, str, None],
    Union[IPv6Address, str],
    Union[IPv6Address, IPv4Address, str, None],
    Union[IPv6Address, IPv4Address, str],
    Union[IPv6Address, IPv4Address],
    Union[IPv6Address, IPv4Address, None],
}


def set_debug() -> None:
    """Sets ENV variable for UFP_DEBUG to on (True)"""
    os.environ[DEBUG_ENV] = str(True)
    is_debug.cache_clear()


def set_no_debug() -> None:
    """Sets ENV variable for UFP_DEBUG to off (False)"""
    os.environ[DEBUG_ENV] = str(False)
    is_debug.cache_clear()


@cache
def is_debug() -> bool:
    """Returns if debug ENV is on (True)"""
    return os.environ.get(DEBUG_ENV) == str(True)


async def get_response_reason(response: ClientResponse) -> str:
    reason = str(response.reason)

    try:
        data = await response.json()
        reason = data.get("error", str(data))
    except Exception:
        with contextlib.suppress(Exception):
            reason = await response.text()

    return reason


@overload
def to_js_time(dt: datetime | int) -> int: ...


@overload
def to_js_time(dt: None) -> None: ...


def to_js_time(dt: datetime | int | None) -> int | None:
    """Converts Python datetime to Javascript timestamp"""
    if dt is None:
        return None

    if isinstance(dt, int):
        return dt

    if dt.tzinfo is None:
        return int(time.mktime(dt.timetuple()) * 1000)

    return int(dt.astimezone(timezone.utc).timestamp() * 1000)


def to_ms(duration: timedelta | None) -> int | None:
    """Converts python timedelta to Milliseconds"""
    if duration is None:
        return None

    return int(round(duration.total_seconds() * 1000))


def utc_now() -> datetime:
    return datetime.now(tz=timezone.utc)


def from_js_time(num: float | str | datetime) -> datetime:
    """Converts Javascript timestamp to Python datetime"""
    if isinstance(num, datetime):
        return num

    return datetime.fromtimestamp(int(num) / 1000, tz=timezone.utc)


@lru_cache(maxsize=1024)
def convert_to_datetime(source_time: float | str | datetime | None) -> datetime | None:
    """Converts timestamp to datetime object"""
    return None if source_time is None else from_js_time(source_time)


def format_datetime(
    dt: datetime | None,
    default: str | None = None,
) -> str | None:
    """Formats a datetime object in a consisent format"""
    return default if dt is None else dt.strftime(DATETIME_FORMAT)


def is_online(data: dict[str, Any]) -> bool:
    return bool(data["state"] == "CONNECTED")


def is_doorbell(data: dict[str, Any]) -> bool:
    return "doorbell" in str(data["type"]).lower()


@lru_cache(maxsize=1024)
def to_snake_case(name: str) -> str:
    """Converts string to snake_case"""
    name = SNAKE_CASE_MATCH_1.sub(r"\1_\2", name)
    name = SNAKE_CASE_MATCH_2.sub(r"_\1", name)
    name = SNAKE_CASE_MATCH_3.sub(r"\1_\2", name)
    return name.lower()


def to_camel_case(name: str) -> str:
    """Converts string to camelCase"""
    # repeated runs through should not keep lowercasing
    if "_" in name:
        name = to_camel(name)
        return name[0].lower() + name[1:]
    return name


_EMPTY_UUID = UUID("0" * 32)
_SHAPE_TYPES = {SHAPE_DICT, SHAPE_LIST, SHAPE_SET}


def convert_unifi_data(value: Any, field: ModelField) -> Any:
    """Converts value from UFP data into pydantic field class"""
    type_ = field.type_

    if type_ is Any:
        return value

    if (shape := field.shape) in _SHAPE_TYPES:
        if shape == SHAPE_LIST and isinstance(value, list):
            return [convert_unifi_data(v, field) for v in value]
        if shape == SHAPE_SET and isinstance(value, list):
            return {convert_unifi_data(v, field) for v in value}
        if shape == SHAPE_DICT and isinstance(value, dict):
            return {k: convert_unifi_data(v, field) for k, v in value.items()}

    if value is not None:
        if type_ in IP_TYPES:
            return _cached_ip_address(value)
        if type_ is datetime:
            return from_js_time(value)
        if type_ in _CREATE_TYPES:
            # cannot do this check too soon because some types cannot be used in isinstance
            if isinstance(value, type_):
                return value
            # handle edge case for improperly formatted UUIDs
            # 00000000-0000-00 0- 000-000000000000
            if type_ is UUID and value == _BAD_UUID:
                return _EMPTY_UUID
            return type_(value)
        if _is_enum_type(type_):
            if _is_from_string_enum(type_):
                return type_.from_string(value)
            return type_(value)

    return value


@lru_cache
def _cached_ip_address(value: str) -> IPv4Address | IPv6Address | str:
    try:
        return ip_address(value)
    except ValueError:
        return value


@lru_cache
def _is_enum_type(type_: Any) -> bool:
    """Checks if type is an Enum."""
    return isclass(type_) and issubclass(type_, Enum)


@lru_cache
def _is_from_string_enum(type_: Any) -> bool:
    """Checks if Enum has from_string method."""
    return hasattr(type_, "from_string")


def serialize_unifi_obj(value: Any, levels: int = -1) -> Any:
    """Serializes UFP data"""
    if unifi_dict := getattr(value, "unifi_dict", None):
        value = unifi_dict()

    if levels != 0 and isinstance(value, dict):
        return serialize_dict(value, levels=levels - 1)
    if levels != 0 and isinstance(value, Iterable) and not isinstance(value, str):
        return serialize_list(value, levels=levels - 1)
    if isinstance(value, Enum):
        return value.value
    if isinstance(value, (IPv4Address, IPv6Address, UUID, Path, tzinfo, Version)):
        return str(value)
    if isinstance(value, datetime):
        return to_js_time(value)
    if isinstance(value, timedelta):
        return to_ms(value)
    if isinstance(value, Color):
        return value.as_hex().upper()

    return value


def serialize_dict(data: dict[str, Any], levels: int = -1) -> dict[str, Any]:
    """Serializes UFP data dict"""
    for key in list(data):
        set_key = key
        if set_key not in SNAKE_CASE_KEYS:
            set_key = to_camel_case(set_key)
        data[set_key] = serialize_unifi_obj(data.pop(key), levels=levels)

    return data


def serialize_coord(coord: CoordType) -> int | float:
    """Serializes UFP zone coordinate"""
    from uiprotect.data import Percent

    if not isinstance(coord, Percent):
        return coord

    if math.isclose(coord, 0) or math.isclose(coord, 1):
        return int(coord)
    return coord


def serialize_point(point: tuple[CoordType, CoordType]) -> list[int | float]:
    """Serializes UFP zone coordinate point"""
    return [
        serialize_coord(point[0]),
        serialize_coord(point[1]),
    ]


def serialize_list(items: Iterable[Any], levels: int = -1) -> list[Any]:
    """Serializes UFP data list"""
    return [serialize_unifi_obj(i, levels=levels) for i in items]


def convert_smart_types(items: Iterable[str]) -> list[SmartDetectObjectType]:
    """Converts list of str into SmartDetectObjectType. Any unknown values will be ignored and logged."""
    types = []
    for smart_type in items:
        try:
            types.append(SmartDetectObjectType(smart_type))
        except ValueError:
            _LOGGER.warning("Unknown smart detect type: %s", smart_type)
    return types


def convert_smart_audio_types(items: Iterable[str]) -> list[SmartDetectAudioType]:
    """Converts list of str into SmartDetectAudioType. Any unknown values will be ignored and logged."""
    types = []
    for smart_type in items:
        try:
            types.append(SmartDetectAudioType(smart_type))
        except ValueError:
            _LOGGER.warning("Unknown smart detect audio type: %s", smart_type)
    return types


def convert_video_modes(items: Iterable[str]) -> list[VideoMode]:
    """Converts list of str into VideoMode. Any unknown values will be ignored and logged."""
    types = []
    for video_mode in items:
        try:
            types.append(VideoMode(video_mode))
        except ValueError:
            _LOGGER.warning("Unknown video mode: %s", video_mode)
    return types


def ip_from_host(host: str) -> IPv4Address | IPv6Address:
    try:
        return ip_address(host)
    except ValueError:
        pass

    return ip_address(socket.gethostbyname(host))


def dict_diff(orig: dict[str, Any] | None, new: dict[str, Any]) -> dict[str, Any]:
    changed: dict[str, Any] = {}

    if orig is None:
        return new

    for key, value in new.items():
        if key not in orig:
            changed[key] = deepcopy(value)
            continue

        if isinstance(value, dict):
            sub_changed = dict_diff(orig[key], value)

            if sub_changed:
                changed[key] = sub_changed
        elif value != orig[key]:
            changed[key] = deepcopy(value)

    return changed


def ws_stat_summmary(
    stats: list[WSStat],
) -> tuple[list[WSStat], float, Counter[str], Counter[str], Counter[str]]:
    if len(stats) == 0:
        raise ValueError("No stats to summarize")

    unfiltered = [s for s in stats if not s.filtered]
    percent = (1 - len(unfiltered) / len(stats)) * 100
    keys = Counter(k for s in unfiltered for k in s.keys_set)
    models = Counter(k.model for k in unfiltered)
    actions = Counter(k.action for k in unfiltered)

    return unfiltered, percent, keys, models, actions


async def write_json(output_path: Path, data: list[Any] | dict[str, Any]) -> None:
    def write() -> None:
        with open(output_path, "w", encoding="utf-8") as f:
            json.dump(data, f, indent=4)
            f.write("\n")

    loop = asyncio.get_running_loop()
    await loop.run_in_executor(None, write)


def print_ws_stat_summary(
    stats: list[WSStat],
    output: Callable[[Any], Any] | None = None,
) -> None:
    # typer<0.4.1 is incompatible with click>=8.1.0
    # allows only the CLI interface to break if both are installed
    import typer

    if output is None:
        output = typer.echo if typer is not None else print

    unfiltered, percent, keys, models, actions = ws_stat_summmary(stats)

    title = " ws stat summary "
    side_length = int((80 - len(title)) / 2)

    lines = [
        "-" * side_length + title + "-" * side_length,
        f"packet count: {len(stats)}",
        f"filtered packet count: {len(unfiltered)} ({percent:.4}%)",
        "-" * 80,
    ]

    for key, count in models.most_common():
        lines.append(f"{key}: {count}")
    lines.append("-" * 80)

    for key, count in actions.most_common():
        lines.append(f"{key}: {count}")
    lines.append("-" * 80)

    for key, count in keys.most_common(10):
        lines.append(f"{key}: {count}")
    lines.append("-" * 80)

    output("\n".join(lines))


async def profile_ws(
    protect: ProtectApiClient,
    duration: int,
    output_path: Path | None = None,
    ws_progress: PROGRESS_CALLABLE | None = None,
    do_print: bool = True,
    print_output: Callable[[Any], Any] | None = None,
) -> None:
    if protect.bootstrap.capture_ws_stats:
        raise NvrError("Profile already in progress")

    _LOGGER.debug("Starting profile...")
    protect.bootstrap.clear_ws_stats()
    protect.bootstrap.capture_ws_stats = True

    if ws_progress is not None:
        await ws_progress(duration, "Waiting for WS messages")
    else:
        await asyncio.sleep(duration)

    protect.bootstrap.capture_ws_stats = False
    _LOGGER.debug("Finished profile...")

    if output_path:
        json_data = [s.__dict__ for s in protect.bootstrap.ws_stats]
        await write_json(output_path, json_data)

    if do_print:
        print_ws_stat_summary(protect.bootstrap.ws_stats, output=print_output)


def decode_token_cookie(token_cookie: Morsel[str]) -> dict[str, Any] | None:
    """Decode a token cookie if it is still valid."""
    try:
        return jwt.decode(
            token_cookie.value,
            options={"verify_signature": False, "verify_exp": True},
        )
    except jwt.ExpiredSignatureError:
        _LOGGER.debug("Authentication token has expired.")
        return None
    except Exception as broad_ex:
        _LOGGER.debug("Authentication token decode error: %s", broad_ex)
        return None


def format_duration(duration: timedelta) -> str:
    """Formats a timedelta as a string."""
    seconds = int(duration.total_seconds())
    hours = seconds // 3600
    seconds -= hours * 3600
    minutes = seconds // 60
    seconds -= minutes * 60

    output = ""
    if hours > 0:
        output = f"{hours}h"
    if minutes > 0:
        output = f"{output}{minutes}m"
    return f"{output}{seconds}s"


def _set_timezone(tz: tzinfo | str) -> tzinfo:
    global TIMEZONE_GLOBAL

    if isinstance(tz, str):
        tz = zoneinfo.ZoneInfo(tz)

    TIMEZONE_GLOBAL = tz

    return TIMEZONE_GLOBAL


def get_local_timezone() -> tzinfo:
    """Gets Olson timezone name for localizing datetimes"""
    if TIMEZONE_GLOBAL is not None:
        return TIMEZONE_GLOBAL

    try:
        from homeassistant.util import dt as dt_util  # type: ignore[import-not-found]

        return _set_timezone(dt_util.DEFAULT_TIME_ZONE)
    except ImportError:
        pass

    timezone_name = os.environ.get("TZ")
    if timezone_name:
        return _set_timezone(timezone_name)

    timezone_name = "UTC"
    timezone_locale = Path("/etc/localtime")
    if timezone_locale.exists():
        tzfile_digest = sha224(Path(timezone_locale).read_bytes()).hexdigest()

        for root, _, filenames in os.walk(Path("/usr/share/zoneinfo/")):
            for filename in filenames:
                fullname = os.path.join(root, filename)
                digest = sha224(Path(fullname).read_bytes()).hexdigest()
                if digest == tzfile_digest:
                    timezone_name = "/".join((fullname.split("/"))[-2:])

    return _set_timezone(timezone_name)


def local_datetime(dt: datetime | None = None) -> datetime:
    """Returns datetime in local timezone"""
    if dt is None:
        dt = datetime.now(tz=timezone.utc)

    local_tz = get_local_timezone()
    if dt.tzinfo is None:
        return dt.replace(tzinfo=local_tz)
    return dt.astimezone(local_tz)


def log_event(event: Event) -> None:
    from uiprotect.data import EventType

    _LOGGER.debug("event WS msg: %s", event.dict())
    if "smart" not in event.type.value:
        return

    camera = event.camera
    if camera is None:
        return

    if event.end is not None:
        _LOGGER.debug(
            "%s (%s): Smart detection ended for %s (%s)",
            camera.name,
            camera.mac,
            event.smart_detect_types,
            event.id,
        )
        return

    _LOGGER.debug(
        "%s (%s): New smart detection started for %s (%s)",
        camera.name,
        camera.mac,
        event.smart_detect_types,
        event.id,
    )
    smart_settings = camera.smart_detect_settings
    for smart_type in event.smart_detect_types:
        is_audio = event.type is EventType.SMART_AUDIO_DETECT
        if is_audio:
            if smart_type.audio_type is None:
                return

            is_enabled = (
                smart_settings.audio_types is not None
                and smart_type.audio_type in smart_settings.audio_types
            )
            last_event = camera.get_last_smart_audio_detect_event(smart_type.audio_type)
        else:
            is_enabled = smart_type in smart_settings.object_types
            last_event = camera.get_last_smart_detect_event(smart_type)

        _LOGGER.debug(
            "Event info (%s):\n"
            "    is_smart_detected: %s\n"
            "    is_recording_enabled: %s\n"
            "    is_enabled: %s\n"
            "    event: %s",
            smart_type,
            camera.is_smart_detected,
            camera.is_recording_enabled,
            is_enabled,
            last_event,
        )


def run_async(callback: Coroutine[Any, Any, T]) -> T:
    """Run async coroutine."""
    if sys.version_info >= (3, 11):
        return asyncio.run(callback)
    loop = asyncio.get_event_loop()  # type: ignore[unreachable]
    return loop.run_until_complete(callback)


def clamp_value(value: float, step_size: float) -> float:
    """Clamps value to multiples of step size."""
    ratio = 1 / step_size
    return int(value * ratio) / ratio


@lru_cache(maxsize=1024)
def normalize_mac(mac: str) -> str:
    """Normalize MAC address."""
    return mac.lower().replace(":", "").replace("-", "").replace("_", "")


_SENTINEL = object()


def get_nested_attr(attrs: tuple[str, ...], obj: Any) -> Any:
    """Fetch a nested attribute."""
    value = obj
    for key in attrs:
        if (value := getattr(value, key, _SENTINEL)) is _SENTINEL:
            return None
    return value


def get_nested_attr_as_bool(attrs: tuple[str, ...], obj: Any) -> bool:
    """Fetch a nested attribute as a bool."""
    value = obj
    for key in attrs:
        if (value := getattr(value, key, _SENTINEL)) is _SENTINEL:
            return False
    return bool(value)


def get_top_level_attr_as_bool(attr: str, obj: Any) -> Any:
    """Fetch a top level attribute as a bool."""
    return bool(getattr(obj, attr))


def make_value_getter(ufp_value: str) -> Callable[[T], Any]:
    """Return a function to get a value from a Protect device."""
    if "." not in ufp_value:
        return attrgetter(ufp_value)
    return partial(get_nested_attr, tuple(ufp_value.split(".")))


def make_enabled_getter(ufp_enabled: str) -> Callable[[T], bool]:
    """Return a function to get a value from a Protect device."""
    if "." not in ufp_enabled:
        return attrgetter(ufp_enabled)
    return partial(get_nested_attr, tuple(ufp_enabled.split(".")))


def make_required_getter(ufp_required_field: str) -> Callable[[T], bool]:
    """Return a function to get a value from a Protect device."""
    if "." not in ufp_required_field:
        return partial(get_top_level_attr_as_bool, ufp_required_field)
    return partial(get_nested_attr_as_bool, tuple(ufp_required_field.split(".")))


@lru_cache
def timedelta_total_seconds(td: timedelta) -> float:
    return td.total_seconds()



from __future__ import annotations

import asyncio
import logging
import shutil
import time
from collections.abc import Callable, Coroutine
from copy import deepcopy
from datetime import datetime, timezone
from pathlib import Path
from shlex import split
from subprocess import run
from typing import Any, overload

import aiohttp
from PIL import Image

from ..api import ProtectApiClient
from ..data import EventType, WSJSONPacketFrame, WSPacket
from ..exceptions import BadRequest
from ..test_util.anonymize import (
    anonymize_data,
    anonymize_prefixed_event_id,
)
from ..utils import from_js_time, is_online, run_async, write_json

BLANK_VIDEO_CMD = "ffmpeg -y -hide_banner -loglevel error -f lavfi -i color=size=1280x720:rate=25:color=black -f lavfi -i anullsrc=channel_layout=stereo:sample_rate=44100 -t {length} {filename}"


def placeholder_image(
    output_path: Path,
    width: int,
    height: int | None = None,
) -> None:
    if height is None:
        height = width

    image = Image.new("RGB", (width, height), (128, 128, 128))
    image.save(output_path, "PNG")


_LOGGER = logging.getLogger(__name__)
LOG_CALLABLE = Callable[[str], None]
PROGRESS_CALLABLE = Callable[[int, str], Coroutine[Any, Any, None]]


class SampleDataGenerator:
    """Generate sample data for debugging and testing purposes"""

    _record_num_ws: int = 0
    _record_ws_start_time: float = time.monotonic()
    _record_listen_for_events: bool = False
    _record_ws_messages: dict[str, dict[str, Any]] = {}
    _log: LOG_CALLABLE | None = None
    _log_warning: LOG_CALLABLE | None = None
    _ws_progress: PROGRESS_CALLABLE | None = None

    constants: dict[str, Any] = {}
    client: ProtectApiClient
    output_folder: Path
    do_zip: bool
    anonymize: bool
    wait_time: int

    def __init__(
        self,
        client: ProtectApiClient,
        output: Path,
        anonymize: bool,
        wait_time: int,
        log: LOG_CALLABLE | None = None,
        log_warning: LOG_CALLABLE | None = None,
        ws_progress: PROGRESS_CALLABLE | None = None,
        do_zip: bool = False,
    ) -> None:
        self.client = client
        self.output_folder = output
        self.do_zip = do_zip
        self.anonymize = anonymize
        self.wait_time = wait_time
        self._log = log
        self._log_warning = log_warning
        self._ws_progress = ws_progress

        if self._log_warning is None and self._log is not None:
            self._log_warning = self._log

    def log(self, msg: str) -> None:
        if self._log is not None:
            self._log(msg)
        else:
            _LOGGER.debug(msg)

    def log_warning(self, msg: str) -> None:
        if self._log_warning is not None:
            self._log_warning(msg)
        else:
            _LOGGER.warning(msg)

    def generate(self) -> None:
        run_async(self.async_generate())

    async def async_generate(self, close_session: bool = True) -> None:
        self.log(f"Output folder: {self.output_folder}")
        self.output_folder.mkdir(parents=True, exist_ok=True)
        websocket = self.client._get_websocket()
        websocket.start()
        self.log("Websocket started...")
        websocket._subscription = self._handle_ws_message

        self.log("Updating devices...")
        await self.client.update()

        bootstrap: dict[str, Any] = await self.client.api_request_obj("bootstrap")
        bootstrap = await self.write_json_file("sample_bootstrap", bootstrap)
        self.constants["server_name"] = bootstrap["nvr"]["name"]
        self.constants["server_id"] = bootstrap["nvr"]["mac"]
        self.constants["server_version"] = bootstrap["nvr"]["version"]
        self.constants["server_ip"] = bootstrap["nvr"]["host"]
        self.constants["server_model"] = bootstrap["nvr"]["type"]
        self.constants["last_update_id"] = bootstrap["lastUpdateId"]
        self.constants["user_id"] = bootstrap["authUserId"]
        self.constants["counts"] = {
            "camera": len(bootstrap["cameras"]),
            "user": len(bootstrap["users"]),
            "group": len(bootstrap["groups"]),
            "liveview": len(bootstrap["liveviews"]),
            "viewer": len(bootstrap["viewers"]),
            "light": len(bootstrap["lights"]),
            "bridge": len(bootstrap["bridges"]),
            "sensor": len(bootstrap["sensors"]),
            "doorlock": len(bootstrap["doorlocks"]),
            "chime": len(bootstrap["chimes"]),
        }

        self.log("Generating event data...")
        motion_event, smart_detection = await self.generate_event_data()
        await self.generate_device_data(motion_event, smart_detection)
        self.log("Recording websocket events...")
        await self.record_ws_events()

        if close_session:
            await self.client.close_session()

        await self.write_json_file("sample_constants", self.constants, anonymize=False)

        if self.do_zip:
            self.log("Zipping files...")

            def zip_files() -> None:
                shutil.make_archive(str(self.output_folder), "zip", self.output_folder)
                shutil.rmtree(self.output_folder)

            loop = asyncio.get_running_loop()
            await loop.run_in_executor(None, zip_files)

    async def record_ws_events(self) -> None:
        if self.wait_time <= 0:
            self.log("Skipping recording Websocket messages...")
            return

        self._record_num_ws = 0
        self._record_ws_start_time = time.monotonic()
        self._record_listen_for_events = True
        self._record_ws_messages = {}

        self.log(f"Waiting {self.wait_time} seconds for WS messages...")
        if self._ws_progress is not None:
            await self._ws_progress(self.wait_time, "Waiting for WS messages")
        else:
            await asyncio.sleep(self.wait_time)

        self._record_listen_for_events = False
        await self.client.async_disconnect_ws()
        await self.write_json_file(
            "sample_ws_messages",
            self._record_ws_messages,
            anonymize=False,
        )

    @overload
    async def write_json_file(
        self,
        name: str,
        data: list[Any],
        anonymize: bool | None = None,
    ) -> list[Any]: ...

    @overload
    async def write_json_file(
        self,
        name: str,
        data: dict[str, Any],
        anonymize: bool | None = None,
    ) -> dict[str, Any]: ...

    async def write_json_file(
        self,
        name: str,
        data: list[Any] | dict[str, Any],
        anonymize: bool | None = None,
    ) -> list[Any] | dict[str, Any]:
        if anonymize is None:
            anonymize = self.anonymize

        if anonymize:
            data = anonymize_data(data)

        self.log(f"Writing {name}...")
        await write_json(self.output_folder / f"{name}.json", data)

        return data

    async def write_binary_file(
        self,
        name: str,
        ext: str,
        raw: bytes | None,
    ) -> None:
        def write() -> None:
            if raw is None:
                self.log(f"No image data, skipping {name}...")
                return

            self.log(f"Writing {name}...")
            Path(self.output_folder / f"{name}.{ext}").write_bytes(raw)

        loop = asyncio.get_running_loop()
        await loop.run_in_executor(None, write)

    async def write_image_file(self, name: str, raw: bytes | None) -> None:
        await self.write_binary_file(name, "png", raw)

    async def generate_event_data(
        self,
    ) -> tuple[dict[str, Any] | None, dict[str, Any] | None]:
        data = await self.client.get_events_raw()

        self.constants["time"] = datetime.now(tz=timezone.utc).isoformat()
        self.constants["event_count"] = len(data)

        motion_event: dict[str, Any] | None = None
        smart_detection: dict[str, Any] | None = None
        for event_dict in reversed(data):
            if (
                motion_event is None
                and event_dict["type"] == EventType.MOTION.value
                and event_dict["camera"] is not None
                and event_dict["thumbnail"] is not None
                and event_dict["heatmap"] is not None
                and event_dict["end"] is not None
            ):
                motion_event = deepcopy(event_dict)
                self.log(f"Using motion event: {motion_event['id']}...")
            elif (
                smart_detection is None
                and event_dict["type"] == EventType.SMART_DETECT.value
                and event_dict["camera"] is not None
                and event_dict["end"] is not None
            ):
                smart_detection = deepcopy(event_dict)
                self.log(f"Using smart detection event: {smart_detection['id']}...")

            if motion_event is not None and smart_detection is not None:
                break

        # anonymize data after pulling events
        data = await self.write_json_file("sample_raw_events", data)

        return motion_event, smart_detection

    async def generate_device_data(
        self,
        motion_event: dict[str, Any] | None,
        smart_detection: dict[str, Any] | None,
    ) -> None:
        await asyncio.gather(
            self.generate_camera_data(),
            self.generate_motion_data(motion_event),
            self.generate_smart_detection_data(smart_detection),
            self.generate_light_data(),
            self.generate_viewport_data(),
            self.generate_sensor_data(),
            self.generate_lock_data(),
            self.generate_chime_data(),
            self.generate_bridge_data(),
            self.generate_liveview_data(),
        )

    async def generate_camera_data(self) -> None:
        objs = await self.client.api_request_list("cameras")
        device_id: str | None = None
        camera_is_online = False
        for obj_dict in objs:
            device_id = obj_dict["id"]
            if is_online(obj_dict):
                camera_is_online = True
                break

        if device_id is None:
            self.log("No camera found. Skipping camera endpoints...")
            return

        # json data
        obj = await self.client.api_request_obj(f"cameras/{device_id}")
        await self.write_json_file("sample_camera", deepcopy(obj))
        self.constants["camera_online"] = camera_is_online

        if not camera_is_online:
            self.log(
                "Camera is not online, skipping snapshot, thumbnail and heatmap generation",
            )

        # snapshot
        width = obj["channels"][0]["width"]
        height = obj["channels"][0]["height"]
        filename = "sample_camera_snapshot"
        if self.anonymize:
            self.log(f"Writing {filename}...")
            placeholder_image(self.output_folder / f"{filename}.png", width, height)
        else:
            snapshot = await self.client.get_camera_snapshot(obj["id"], width, height)
            await self.write_image_file(filename, snapshot)

    async def generate_motion_data(
        self,
        motion_event: dict[str, Any] | None,
    ) -> None:
        if motion_event is None:
            self.log("No motion event, skipping thumbnail and heatmap generation...")
            return

        # event thumbnail
        filename = "sample_camera_thumbnail"
        thumbnail_id = motion_event["thumbnail"]
        if self.anonymize:
            self.log(f"Writing {filename}...")
            placeholder_image(self.output_folder / f"{filename}.png", 640, 360)
            thumbnail_id = anonymize_prefixed_event_id(thumbnail_id)
        else:
            img = await self.client.get_event_thumbnail(thumbnail_id)
            await self.write_image_file(filename, img)
        self.constants["camera_thumbnail"] = thumbnail_id

        # event heatmap
        filename = "sample_camera_heatmap"
        heatmap_id = motion_event["heatmap"]
        if self.anonymize:
            self.log(f"Writing {filename}...")
            placeholder_image(self.output_folder / f"{filename}.png", 640, 360)
            heatmap_id = anonymize_prefixed_event_id(heatmap_id)
        else:
            img = await self.client.get_event_heatmap(heatmap_id)
            await self.write_image_file(filename, img)
        self.constants["camera_heatmap"] = heatmap_id

        # event video
        filename = "sample_camera_video"
        length = int((motion_event["end"] - motion_event["start"]) / 1000)
        if self.anonymize:
            run(
                split(
                    BLANK_VIDEO_CMD.format(
                        length=length,
                        filename=self.output_folder / f"{filename}.mp4",
                    ),
                ),
                check=True,
            )
        else:
            video = await self.client.get_camera_video(
                motion_event["camera"],
                from_js_time(motion_event["start"]),
                from_js_time(motion_event["end"]),
                2,
            )
            await self.write_binary_file(filename, "mp4", video)
        self.constants["camera_video_length"] = length

    async def generate_smart_detection_data(
        self,
        smart_detection: dict[str, Any] | None,
    ) -> None:
        if smart_detection is None:
            self.log("No smart detection event, skipping smart detection data...")
            return

        try:
            data = await self.client.get_event_smart_detect_track_raw(
                smart_detection["id"],
            )
        except BadRequest:
            self.log_warning("Event smart tracking missing")
        else:
            await self.write_json_file("sample_event_smart_track", data)

    async def generate_light_data(self) -> None:
        objs = await self.client.api_request_list("lights")
        device_id: str | None = None
        for obj_dict in objs:
            device_id = obj_dict["id"]
            if is_online(obj_dict):
                break

        if device_id is None:
            self.log("No light found. Skipping light endpoints...")
            return

        obj = await self.client.api_request_obj(f"lights/{device_id}")
        await self.write_json_file("sample_light", obj)

    async def generate_viewport_data(self) -> None:
        objs = await self.client.api_request_list("viewers")
        device_id: str | None = None
        for obj_dict in objs:
            device_id = obj_dict["id"]
            if is_online(obj_dict):
                break

        if device_id is None:
            self.log("No viewer found. Skipping viewer endpoints...")
            return

        obj = await self.client.api_request_obj(f"viewers/{device_id}")
        await self.write_json_file("sample_viewport", obj)

    async def generate_sensor_data(self) -> None:
        objs = await self.client.api_request_list("sensors")
        device_id: str | None = None
        for obj_dict in objs:
            device_id = obj_dict["id"]
            if is_online(obj_dict):
                break

        if device_id is None:
            self.log("No sensor found. Skipping sensor endpoints...")
            return

        obj = await self.client.api_request_obj(f"sensors/{device_id}")
        await self.write_json_file("sample_sensor", obj)

    async def generate_lock_data(self) -> None:
        objs = await self.client.api_request_list("doorlocks")
        device_id: str | None = None
        for obj_dict in objs:
            device_id = obj_dict["id"]
            if is_online(obj_dict):
                break

        if device_id is None:
            self.log("No doorlock found. Skipping doorlock endpoints...")
            return

        obj = await self.client.api_request_obj(f"doorlocks/{device_id}")
        await self.write_json_file("sample_doorlock", obj)

    async def generate_chime_data(self) -> None:
        objs = await self.client.api_request_list("chimes")
        device_id: str | None = None
        for obj_dict in objs:
            device_id = obj_dict["id"]
            if is_online(obj_dict):
                break

        if device_id is None:
            self.log("No chime found. Skipping doorlock endpoints...")
            return

        obj = await self.client.api_request_obj(f"chimes/{device_id}")
        await self.write_json_file("sample_chime", obj)

    async def generate_bridge_data(self) -> None:
        objs = await self.client.api_request_list("bridges")
        device_id: str | None = None
        for obj_dict in objs:
            device_id = obj_dict["id"]
            if is_online(obj_dict):
                break

        if device_id is None:
            self.log("No bridge found. Skipping bridge endpoints...")
            return

        obj = await self.client.api_request_obj(f"bridges/{device_id}")
        await self.write_json_file("sample_bridge", obj)

    async def generate_liveview_data(self) -> None:
        objs = await self.client.api_request_list("liveviews")
        device_id: str | None = None
        for obj_dict in objs:
            device_id = obj_dict["id"]
            break

        if device_id is None:
            self.log("No liveview found. Skipping liveview endpoints...")
            return

        obj = await self.client.api_request_obj(f"liveviews/{device_id}")
        await self.write_json_file("sample_liveview", obj)

    def _handle_ws_message(self, msg: aiohttp.WSMessage) -> None:
        if not self._record_listen_for_events:
            return

        now = time.monotonic()
        self._record_num_ws += 1
        time_offset = now - self._record_ws_start_time

        if msg.type == aiohttp.WSMsgType.BINARY:
            packet = WSPacket(msg.data)

            if not isinstance(packet.action_frame, WSJSONPacketFrame):
                self.log_warning(
                    f"Got non-JSON action frame: {packet.action_frame.payload_format}",
                )
                return

            if not isinstance(packet.data_frame, WSJSONPacketFrame):
                self.log_warning(
                    f"Got non-JSON data frame: {packet.data_frame.payload_format}",
                )
                return

            if self.anonymize:
                packet.action_frame.data = anonymize_data(packet.action_frame.data)
                packet.data_frame.data = anonymize_data(packet.data_frame.data)
                packet.pack_frames()

            self._record_ws_messages[str(time_offset)] = {
                "raw": packet.raw_base64,
                "action": packet.action_frame.data,
                "data": packet.data_frame.data,
            }
        else:
            self.log_warning(f"Got non-binary message: {msg.type}")
